# find.py
# Copyright (c) 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""A record selection statement evaluator approximately equivalent to SQL
Select statement where clause and DPT Find statement.

Classes:

Find


Notes:

The statement syntax is defined in where.py module docstring.

"""

import re

from . import where
from .recordset import Recordset


class Find():

    """Selection statement evaluator for a Database instance primary table.

    The methods of the Database instance db are used to evaluate the request on
    the primary table named in dbset.
    
    """

    def __init__(self, db, dbset, recordclass=None):
        self._db = db
        self._dbset = dbset
        self._recordclass = recordclass
        # True means Alpha test, False means Num, None means neither.
        # In particular f eq v defaults to f alpha eq v but f is v can be
        # seen as f eq v without the defaulting.
        # Alpha: b>a b>ab 2>1 2>12
        # Num:   b>a ab>b 2>1 12>2
        # In this test Num works for integer numbers only.
        self.compare_field_value = {
            (where.IS, None): self._is,
            (where.LIKE, None): self._like,
            (where.PRESENT, None): self._present,
            (where.EQ, True): self._eq,
            (where.NE, True): self._ne,
            (where.GT, True): self._gt,
            (where.LT, True): self._lt,
            (where.LE, True): self._le,
            (where.GE, True): self._ge,
            (where.BEFORE, True): self._before,
            (where.AFTER, True): self._after,
            ((where.FROM, where.TO), True): self._from_to,
            ((where.FROM, where.BELOW), True): self._from_below,
            ((where.ABOVE, where.TO), True): self._above_to,
            ((where.ABOVE, where.BELOW), True): self._above_below,
            (where.EQ, False): self._eq,
            (where.NE, False): self._ne,
            (where.GT, False): self._gt,
            (where.LT, False): self._lt,
            (where.LE, False): self._le,
            (where.GE, False): self._ge,
            (where.BEFORE, False): self._before,
            (where.AFTER, False): self._after,
            ((where.FROM, where.TO), False): self._from_to,
            ((where.FROM, where.BELOW), False): self._from_below,
            ((where.ABOVE, where.TO), False): self._above_to,
            ((where.ABOVE, where.BELOW), False): self._above_below,
            }
        self.boolean_operation = {
            where.AND: self._and,
            where.NOR: self._nor,
            where.OR: self._or,
            }
        
    def condition(self, obj):
        if not self._db.exists(self._dbset, obj.field):
            return None
        if obj.condition in {where.IS, where.LIKE, where.PRESENT}:
            case = (obj.condition, None)
        elif obj.num is True:
            case = (obj.condition, False)
        else:
            case = (obj.condition, True)
        recordset = self.compare_field_value[case](obj)
        # Let parser stop 'field not is not value' 'field not is value'
        # 'field is not value' is only case of 'field <condition> not value'
        if obj.not_condition:
            recordset = self.get_existence() - recordset
        if obj.not_phrase:
            recordset = self.get_existence() - recordset
        obj.result.answer = recordset
        
    def not_condition(self, obj):
        if obj.not_condition:
            obj.result.answer = self.get_existence() - obj.result.answer
        if obj.not_phrase:
            obj.result.answer = self.get_existence() - obj.result.answer
    
    def operator(self, obj):
        obj.left.result.answer = self.boolean_operation[obj.operator](obj)
        obj.result = obj.left.result
    
    def answer(self, obj):
        # Evaluate 'not' in clauses like '... and not (...) ...'
        if obj.not_phrase:
            obj.result.answer = self.get_existence() - obj.result.answer
        obj.up.result = obj.result
    
    def initialize_answer(self, obj):
        obj.result.answer = Recordset(self._db, self._dbset)

    def get_existence(self):
        return self._db.make_recordset_all(self._dbset, self._dbset)
    
    def get_record(self, recordset):
        instance = self._recordclass()
        for segment in recordset.rs_segments.values():
            rp = segment.first()
            while rp:
                instance.load_record(
                    self._db.get_primary_record(self._dbset, rp[1]))
                yield rp[1], instance.value
                rp = segment.next()
        
    def non_index_condition(self, obj, record_number, record):
        if obj.condition == where.IS:
            if obj.not_value:
                self._is_not(obj, record_number, record)
        elif obj.condition == where.NE:
            self._ne(obj, record_number, record)
        elif obj.condition == where.PRESENT:
            self._present(obj, record_number, record)
        elif obj.condition == where.LIKE:
            self._like(obj, record_number, record)

    def _is(self, obj):
        # 'field is value' and 'field is not value' are allowed
        if obj.not_value:
            raise RuntimeError('_is')
        else:
            return self._db.make_recordset_key(
                self._dbset,
                obj.field,
                key=self._db.encode_record_selector(obj.value))

    def _is_not(self, obj, record_number, record):
        # 'field is value' and 'field is not value' are allowed
        if obj.not_value:
            f = record.get_field_values(obj.field)
            if f:
                for v in f:
                    if v != obj.value:
                        obj.result.answer.place(record_number)
                        break
        else:
            raise RuntimeError('_is_not')

    def _like(self, obj, record_number, record):
        f = record.get_field_values(obj.field)
        if f:
            for v in f:
                try:
                    if re.search(obj.value, v):
                        obj.result.answer.place(record_number)
                        break
                except:
                    pass

    def _present(self, obj, record_number, record):
        if record.get_field_values(obj.field):
            obj.result.answer.place(record_number)

    def _eq(self, obj):
        return self._db.make_recordset_key(
            self._dbset,
            obj.field,
            key=self._db.encode_record_selector(obj.value))

    def _ne(self, obj, record_number, record):
        f = record.get_field_values(obj.field)
        if f:
            for v in f:
                if v != obj.value:
                    obj.result.answer.place(record_number)
                    break

    def _gt(self, obj):
        gt = self._gt_key(obj)
        if gt is None:
            return Recordset(self._db, self._dbset)
        return self._db.make_recordset_key_range(
            self._dbset, obj.field, key=gt)

    def _lt(self, obj):
        lt = self._lt_key(obj)
        return self._db.make_recordset_key_range(
            self._dbset, obj.field, keyend=lt)

    def _le(self, obj):
        return self._db.make_recordset_key_range(
            self._dbset,
            obj.field,
            keyend=self._db.encode_record_selector(obj.value))

    def _ge(self, obj):
        return self._db.make_recordset_key_range(
            self._dbset,
            obj.field,
            key=self._db.encode_record_selector(obj.value))

    def _before(self, obj):
        return self._db.make_recordset_key_range(
            self._dbset, obj.field, keyend=self._before_key(obj))

    def _after(self, obj):
        after = self._after_key(obj)
        if after is None:
            return Recordset(self._db, self._dbset)
        return self._db.make_recordset_key_range(
            self._dbset, obj.field, key=after)

    def _from_to(self, obj):
        return self._db.make_recordset_key_range(
            self._dbset,
            obj.field,
            key=self._db.encode_record_selector(obj.value[0]),
            keyend=self._db.encode_record_selector(obj.value[1]))

    def _from_below(self, obj):
        return self._db.make_recordset_key_range(
            self._dbset,
            obj.field,
            key=self._db.encode_record_selector(obj.value[0]),
            keyend=self._below_key(obj))

    def _above_to(self, obj):
        above = self._above_key(obj)
        if above is None:
            return Recordset(self._db, self._dbset)
        return self._db.make_recordset_key_range(
            self._dbset,
            obj.field,
            key=above,
            keyend=self._db.encode_record_selector(obj.value[1]))

    def _above_below(self, obj):
        above = self._above_key(obj)
        if above is None:
            return Recordset(self._db, self._dbset)
        return self._db.make_recordset_key_range(
            self._dbset,
            obj.field,
            key=above,
            keyend=self._below_key(obj))

    def _and(self, obj):
        return obj.left.result.answer & obj.result.answer

    def _nor(self, obj):
        return obj.left.result.answer & (
            self.get_existence() - obj.result.answer)

    def _or(self, obj):
        return obj.left.result.answer | obj.result.answer

    def _gt_key(self, obj):
        c = self._db.database_cursor(self._dbset, obj.field)
        r = c.nearest(self._db.encode_record_selector(obj.value))
        if r:
            k = r[0][0]
            if k > obj.value:
                return k
            r = c.next()
            k = r[0][0]
            if k > obj.value:
                return k
            

    def _lt_key(self, obj):
        c = self._db.database_cursor(self._dbset, obj.field)
        r = c.nearest(self._db.encode_record_selector(obj.value))
        if r:
            k = r[0][0]
            if k < obj.value:
                return k
            r = c.prev()
            k = r[0][0]
            if k < obj.value:
                return k

    def _before_key(self, obj):
        return self._lt_key(obj)

    def _after_key(self, obj):
        return self._gt_key(obj)

    def _below_key(self, obj):
        return self._lt_key(obj)

    def _above_key(self, obj):
        return self._gt_key(obj)
