# _filespec.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Sample file specification for use in tests"""

import shutil
import os

try:
    from bsddb3.db import (
        DB_BTREE, DB_HASH, DB_RECNO, DB_DUPSORT, DB_DUP,
        DB_CREATE, DB_INIT_MPOOL, DB_PRIVATE,
        )
except ImportError:
    pass

from ..constants import (
    PRIMARY,
    SECONDARY,
    DDNAME,
    FILE,
    FIELDS,
    FILEDESC,
    INV,
    ORD,
    RRN, 
    DB_UPPER_CONVERSION_LIMIT,
    DB_LOWER_CONVERSION_LIMIT,
    )
from .. import filespec
from ..record import KeyData, Value, Record
from ...sqlite3api import Sqlite3bitapi, Sqlite3apiError
try:
    from ...dbapi import DBbitapi, DBapiError
except ImportError:
    pass

GAMES_FILE_DEF = 'games'
GAME_FIELD_DEF = 'Game'
EVENT_FIELD_DEF = 'Event'
SITE_FIELD_DEF = 'Site'
DATE_FIELD_DEF = 'Date'
ROUND_FIELD_DEF = 'Round'
WHITE_FIELD_DEF = 'White'
BLACK_FIELD_DEF = 'Black'
RESULT_FIELD_DEF = 'Result'
SEVEN_TAG_ROSTER = (EVENT_FIELD_DEF,
                    SITE_FIELD_DEF,
                    DATE_FIELD_DEF,
                    ROUND_FIELD_DEF,
                    WHITE_FIELD_DEF,
                    BLACK_FIELD_DEF,
                    RESULT_FIELD_DEF,
                    )
NAME_FIELD_DEF = 'Name'


def samplefilespec():
    """Return a FileSpec for a sample database"""
    class FileSpec(filespec.FileSpec):
        def __init__(self):
            dptdsn = FileSpec.dpt_dsn
            fn = FileSpec.field_name
            super().__init__(**{
                GAMES_FILE_DEF: {
                    DDNAME: 'GAMES',
                    FILE: dptdsn(GAMES_FILE_DEF),
                    PRIMARY: fn(GAME_FIELD_DEF),
                    SECONDARY : {
                        EVENT_FIELD_DEF: None,
                        SITE_FIELD_DEF: None,
                        DATE_FIELD_DEF: None,
                        ROUND_FIELD_DEF: None,
                        WHITE_FIELD_DEF: None,
                        BLACK_FIELD_DEF: None,
                        RESULT_FIELD_DEF: None,
                        NAME_FIELD_DEF: None,
                        },
                    FIELDS: {
                        GAME_FIELD_DEF: None,
                        WHITE_FIELD_DEF: {INV:True, ORD:True},
                        BLACK_FIELD_DEF: {INV:True, ORD:True},
                        EVENT_FIELD_DEF: {INV:True, ORD:True},
                        ROUND_FIELD_DEF: {INV:True, ORD:True},
                        DATE_FIELD_DEF: {INV:True, ORD:True},
                        RESULT_FIELD_DEF: {INV:True, ORD:True},
                        SITE_FIELD_DEF: {INV:True, ORD:True},
                        NAME_FIELD_DEF: {INV:True, ORD:True},
                        },
                    },
                })
    return FileSpec()
        

class SampleValue(Value):

    def pack(self):
        v = super().pack()
        index = v[1]
        for field in SEVEN_TAG_ROSTER:
            if hasattr(self, field):
                index[field] = [getattr(self, field)]
        return v


class SampleRecord(Record):

    def __init__(self):
        super().__init__(KeyData, SampleValue)
        
    def get_keys(self, datasource=None, partial=None):
        dbname = datasource.dbname
        if dbname == GAMES_FILE_DEF:
            return [(self.key.recno, self.srvalue)]
        elif dbname in SEVEN_TAG_ROSTER:
            if hasattr(self.value, field):
                return [(getattr(self.value, dbname), self.key.pack())]
        return []
        

class SampleNameValue(SampleValue):

    def pack(self):
        v = super().pack()
        name = []
        for field in (WHITE_FIELD_DEF, BLACK_FIELD_DEF):
            if hasattr(self, field):
                name.append(getattr(self, field))
        if len(name):
            v[1][NAME_FIELD_DEF] = name
        return v


class SampleNameRecord(Record):

    def __init__(self):
        super().__init__(KeyData, SampleNameValue)
        
    def get_keys(self, datasource=None, partial=None):
        dbname = datasource.dbname
        if dbname == GAMES_FILE_DEF:
            return [(self.key.recno, self.srvalue)]
        elif dbname in SEVEN_TAG_ROSTER:
            if hasattr(self.value, field):
                return [(getattr(self.value, dbname), self.key.pack())]
        return []


class Sqlite3Database(Sqlite3bitapi):

    def __init__(self, sqlite3file):
        names = samplefilespec()
        try:
            super().__init__(
                names,
                sqlite3file,
                )
        except Sqlite3apiError:
            if __name__ == '__main__':
                raise
            else:
                raise Sqlite3apiError('sqlite3 description invalid')

    def delete_database(self):
        names = {self._sqfile}
        basenames = set()
        listnames = []
        self.close_database()
        for n in names:
            if os.path.isdir(n):
                shutil.rmtree(n, ignore_errors=True)
            else:
                os.remove(n)
        try:
            d, f = os.path.split(self._sqfile)
            if f == os.path.basename(d):
                os.rmdir(d)
        except:
            pass
        return True


try:
    class Bsddb3Database(DBbitapi):

        def __init__(self, DBfile):
            names = samplefilespec()
            try:
                super().__init__(
                    names,
                    DBfile,
                    DBenvironment={'flags': DB_CREATE | DB_INIT_MPOOL},
                    )
            except DBapiError:
                if __name__ == '__main__':
                    raise
                else:
                    raise DBapiError('sqlite3 description invalid')

        def delete_database(self):
            files = []
            folders = []
            for k, v in self._main.items():
                files.append(os.path.join(
                    self.get_database_folder(), v.get_database_file()))
                if not v.is_primary():
                    df = os.path.join(
                        self.get_database_folder(),
                        self.get_deferred_update_folder(),
                        v.get_database_file())
                    if os.path.isdir(df):
                        folders.append(df)
                        for n in os.listdir(df):
                            if (n.startswith('shelf') and
                                n[len('shelf'):].isdigit()):
                                files.append(os.path.join(df, n))
            folders.append(os.path.join(
                self.get_database_folder(), self.get_deferred_update_folder()))
            folders.append(self.get_database_folder())
            self.close_database()
            for f in files:
                if os.path.isfile(f):
                    try:
                        os.remove(f)
                    except:
                        pass
            for f in folders:
                if os.path.isdir(f):
                    try:
                        os.rmdir(f)
                    except:
                        pass
            return True


except NameError:
    # Assume bsddb3 is not installed
    pass
