# find_test.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""find tests"""

import unittest
from copy import copy, deepcopy
import os
from ast import literal_eval

from ._filespec import (GAMES_FILE_DEF,
                        GAME_FIELD_DEF,
                        EVENT_FIELD_DEF,
                        SITE_FIELD_DEF,
                        DATE_FIELD_DEF,
                        ROUND_FIELD_DEF,
                        WHITE_FIELD_DEF,
                        BLACK_FIELD_DEF,
                        RESULT_FIELD_DEF,
                        SEVEN_TAG_ROSTER,
                        NAME_FIELD_DEF,
                        samplefilespec,
                        Sqlite3Database,
                        SampleNameValue,
                        SampleNameRecord,
                        )

from .. import where
from .. import find


class FindTC(unittest.TestCase):

    def setUp(self):
        self.findtest_database = os.path.expanduser(
            os.path.join('~', '_findtest_database'))
        self.filespec = samplefilespec()
        self.findtest_record = SampleNameRecord()
        # Usually data[index] = function(data[GAME_FIELD_DEF]) where index is
        # one of the *_FIELD_DEF items.
        # For testing just set some arbitrary plausible values.
        self.testdata = (
            {GAME_FIELD_DEF:'gamedata1',
             EVENT_FIELD_DEF:'eventdata2',
             SITE_FIELD_DEF:'sitedata3',
             DATE_FIELD_DEF:'datedata1',
             ROUND_FIELD_DEF:'rounddata2',
             WHITE_FIELD_DEF:'whitedata3',
             BLACK_FIELD_DEF:'blackdata0',
             RESULT_FIELD_DEF:'resultdata0',
             },
            {GAME_FIELD_DEF:'gamedata2',
             EVENT_FIELD_DEF:'eventdata3',
             SITE_FIELD_DEF:'sitedata1',
             DATE_FIELD_DEF:'datedata3',
             ROUND_FIELD_DEF:'rounddata1',
             WHITE_FIELD_DEF:'whitedata2',
             BLACK_FIELD_DEF:'blackdata0',
             RESULT_FIELD_DEF:'resultdata0',
             },
            {GAME_FIELD_DEF:'gamedata3',
             EVENT_FIELD_DEF:'eventdata1',
             SITE_FIELD_DEF:'sitedata2',
             DATE_FIELD_DEF:'datedata1',
             ROUND_FIELD_DEF:'rounddata2',
             WHITE_FIELD_DEF:'whitedata3',
             BLACK_FIELD_DEF:'blackdata0',
             RESULT_FIELD_DEF:'resultdata0',
             },
            )
        self.trdata = {GAME_FIELD_DEF:'game',
                       BLACK_FIELD_DEF:'black',
                       RESULT_FIELD_DEF:'result',
                       }

    def tearDown(self):
        pass

    def check_index(self, database, field, expected_records):
        c = database.database_cursor(GAMES_FILE_DEF, field)
        r = c.first()
        for er in expected_records:
            self.assertEqual(r, er)
            r = c.next()
        else:
            self.assertEqual(r, None)

    def check_table(self, database, expected_records):
        c = database.database_cursor(GAMES_FILE_DEF, GAMES_FILE_DEF)
        r = c.first()
        for er in expected_records:
            self.assertEqual((r[0], literal_eval(r[1])), er)
            r = c.next()
        else:
            self.assertEqual(r, None)

    def test___raises(self):
        """"""
        pass

    def test___copy(self):
        """"""
        pass

    def test____assumptions(self):
        """"""
        msg = 'Failure of this test invalidates all other tests'
        ftd = Sqlite3Database(self.findtest_database)
        try:
            if ftd.open_context():
                for r in self.testdata:
                    ftd.start_transaction()
                    self.findtest_record.load_record((None, repr(r)))
                    self.findtest_record.put_record(ftd, GAMES_FILE_DEF)
                    ftd.commit()
                rd = {k:''.join((v, str(4).zfill(4)))
                      for k, v in self.trdata.items()}
                ftd.start_transaction()
                for r in range(10):
                    self.findtest_record.load_record((None, repr(rd)))
                    self.findtest_record.put_record(ftd, GAMES_FILE_DEF)
                ftd.commit()
                self.check_index(
                    ftd,
                    RESULT_FIELD_DEF,
                    [('result0004', 4),
                     ('result0004', 5),
                     ('result0004', 6),
                     ('result0004', 7),
                     ('result0004', 8),
                     ('result0004', 9),
                     ('result0004', 10),
                     ('result0004', 11),
                     ('result0004', 12),
                     ('result0004', 13),
                     ('resultdata0', 1),
                     ('resultdata0', 2),
                     ('resultdata0', 3),
                     ])
                self.check_index(
                    ftd,
                    BLACK_FIELD_DEF,
                    [('black0004', 4),
                     ('black0004', 5),
                     ('black0004', 6),
                     ('black0004', 7),
                     ('black0004', 8),
                     ('black0004', 9),
                     ('black0004', 10),
                     ('black0004', 11),
                     ('black0004', 12),
                     ('black0004', 13),
                     ('blackdata0', 1),
                     ('blackdata0', 2),
                     ('blackdata0', 3),
                     ])
                self.check_index(
                    ftd,
                    EVENT_FIELD_DEF,
                    [('eventdata1', 3),
                     ('eventdata2', 1),
                     ('eventdata3', 2),
                     ])
                self.check_index(
                    ftd,
                    SITE_FIELD_DEF,
                    [('sitedata1', 2),
                     ('sitedata2', 3),
                     ('sitedata3', 1),
                     ])
                self.check_index(
                    ftd,
                    DATE_FIELD_DEF,
                    [('datedata1', 1),
                     ('datedata1', 3),
                     ('datedata3', 2),
                     ])
                self.check_index(
                    ftd,
                    ROUND_FIELD_DEF,
                    [('rounddata1', 2),
                     ('rounddata2', 1),
                     ('rounddata2', 3),
                     ])
                self.check_index(
                    ftd,
                    WHITE_FIELD_DEF,
                    [('whitedata2', 2),
                     ('whitedata3', 1),
                     ('whitedata3', 3),
                     ])
                self.check_index(
                    ftd,
                    NAME_FIELD_DEF,
                    [('black0004', 4),
                     ('black0004', 5),
                     ('black0004', 6),
                     ('black0004', 7),
                     ('black0004', 8),
                     ('black0004', 9),
                     ('black0004', 10),
                     ('black0004', 11),
                     ('black0004', 12),
                     ('black0004', 13),
                     ('blackdata0', 1),
                     ('blackdata0', 2),
                     ('blackdata0', 3),
                     ('whitedata2', 2),
                     ('whitedata3', 1),
                     ('whitedata3', 3),
                     ])
                self.check_table(
                    ftd,
                    [(1, {'Event': 'eventdata2',
                          'Date': 'datedata1',
                          'Round': 'rounddata2',
                          'White': 'whitedata3',
                          'Site': 'sitedata3',
                          'Game':'gamedata1',
                          'Result':'resultdata0',
                          'Black':'blackdata0'}),
                     (2, {'Event': 'eventdata3',
                          'Date': 'datedata3',
                          'Round': 'rounddata1',
                          'White': 'whitedata2',
                          'Site': 'sitedata1',
                          'Game':'gamedata2',
                          'Result':'resultdata0',
                          'Black':'blackdata0'}),
                     (3, {'Event': 'eventdata1',
                          'Date': 'datedata1',
                          'Round': 'rounddata2',
                          'White': 'whitedata3',
                          'Site': 'sitedata2',
                          'Game':'gamedata3',
                          'Result':'resultdata0',
                          'Black':'blackdata0'}),
                     (4, {'Game':'game0004',
                          'Result':'result0004',
                          'Black':'black0004'}),
                     (5, {'Game':'game0004',
                          'Result':'result0004',
                          'Black':'black0004'}),
                     (6, {'Game':'game0004',
                          'Result':'result0004',
                          'Black':'black0004'}),
                     (7, {'Game':'game0004',
                          'Result':'result0004',
                          'Black':'black0004'}),
                     (8, {'Game':'game0004',
                          'Result':'result0004',
                          'Black':'black0004'}),
                     (9, {'Game':'game0004',
                          'Result':'result0004',
                          'Black':'black0004'}),
                     (10, {'Game':'game0004',
                           'Result':'result0004',
                           'Black':'black0004'}),
                     (11, {'Game':'game0004',
                           'Result':'result0004',
                           'Black':'black0004'}),
                     (12, {'Game':'game0004',
                           'Result':'result0004',
                           'Black':'black0004'}),
                     (13, {'Game':'game0004',
                           'Result':'result0004',
                           'Black':'black0004'}),
                     ])
        finally:
            ftd.delete_database()

    def test____module_constants(self):
        """"""

    def test___init__(self):
        """"""
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 2 required positional arguments: ",
                "'db' and 'dbset'",
                )),
            find.Find,
            )
        f = find.Find(True, False)
        self.assertEqual(len(f.__dict__), 4)
        self.assertEqual(f._db, True)
        self.assertEqual(f._dbset, False)
        self.assertEqual(
            f.compare_field_value,
            {(where.IS, None): f._is,
             (where.LIKE, None): f._like,
             (where.PRESENT, None): f._present,
             (where.EQ, True): f._eq,
             (where.NE, True): f._ne,
             (where.GT, True): f._gt,
             (where.LT, True): f._lt,
             (where.LE, True): f._le,
             (where.GE, True): f._ge,
             (where.BEFORE, True): f._before,
             (where.AFTER, True): f._after,
             ((where.FROM, where.TO), True): f._from_to,
             ((where.FROM, where.BELOW), True): f._from_below,
             ((where.ABOVE, where.TO), True): f._above_to,
             ((where.ABOVE, where.BELOW), True): f._above_below,
             (where.EQ, False): f._eq,
             (where.NE, False): f._ne,
             (where.GT, False): f._gt,
             (where.LT, False): f._lt,
             (where.LE, False): f._le,
             (where.GE, False): f._ge,
             (where.BEFORE, False): f._before,
             (where.AFTER, False): f._after,
             ((where.FROM, where.TO), False): f._from_to,
             ((where.FROM, where.BELOW), False): f._from_below,
             ((where.ABOVE, where.TO), False): f._above_to,
             ((where.ABOVE, where.BELOW), False): f._above_below,
             })
        self.assertEqual(
            f.boolean_operation,
            {where.AND: f._and,
             where.NOR: f._nor,
             where.OR: f._or,
             })
        

if __name__ == '__main__':
    runner = unittest.TextTestRunner
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    runner().run(loader(FindTC))
