# where.py
# Copyright (c) 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""A record selection statement parser approximately equivalent to SQL Select
statement where clause and DPT Find statement retrieval conditions.

Classes:

Where
WhereClause
WhereConstraint
WhereError
WhereResult

Functions:

_trim


The syntax is:

[NOT] phrase [<AND|OR|NOR> [NOT] phrase] ...

where phrase is one of:

fieldname IS [NOT] value
fieldname [NOT] LIKE pattern
fieldname [NOT] PRESENT
fieldname [NOT] [NUM|ALPHA] <EQ|NE|GT|LT|LE|GE|BEFORE|AFTER> value
fieldname [NOT] [NUM|ALPHA] <FROM|ABOVE> value <TO|BELOW> value

| indicates choices.
[] indicates optional items and <> indicates choice in non-optional items.

Priority of operators is, in decreasing order: NOT NOR AND OR.

Parentheses can be placed around a sequence of phrases to override the normal
priority order.  The phrases within the parentheses are evaluated first.

It is possible for "fieldname IS NOT value" to give a different answer than
"fieldname NOT EQ value" or "fieldname NE value" or "NOT fieldmae EQ value",
particularly the last of these.

"""

import re

DOUBLE_QUOTE_STRING = '".*?"'
SINGLE_QUOTE_STRING = "'.*?'"
LEFT_PARENTHESIS = '('
RIGHT_PARENTHESIS = ')'
OR = 'or'
TO = 'to'
IS = 'is'
EQ = 'eq'
NE = 'ne'
LT = 'lt'
LE = 'le'
GT = 'gt'
GE = 'ge'
NOT = 'not'
NOR = 'nor'
AND = 'and'
NUM = 'num'
LIKE = 'like'
FROM = 'from'
ALPHA = 'alpha'
ABOVE = 'above'
AFTER = 'after'
BELOW = 'below'
BEFORE = 'before'
PRESENT = 'present'
STRING = '.*?'

LEADING_SPACE = '(?<=\s)'
TRAILING_SPACE = '(?=\s)'

WHERE_RE = re.compile(
    '|'.join((DOUBLE_QUOTE_STRING,
              SINGLE_QUOTE_STRING,
              ''.join(('\\', LEFT_PARENTHESIS)),
              ''.join(('\\', RIGHT_PARENTHESIS)),
              OR.join(('(?<=\s|\))', '(?=\s|\()')),
              TO.join((LEADING_SPACE, TRAILING_SPACE)),
              IS.join((LEADING_SPACE, TRAILING_SPACE)),
              EQ.join((LEADING_SPACE, TRAILING_SPACE)),
              NE.join((LEADING_SPACE, TRAILING_SPACE)),
              LT.join((LEADING_SPACE, TRAILING_SPACE)),
              LE.join((LEADING_SPACE, TRAILING_SPACE)),
              GT.join((LEADING_SPACE, TRAILING_SPACE)),
              GE.join((LEADING_SPACE, TRAILING_SPACE)),
              NOT.join(('\A', '(?=\s|\()')),
              NOT.join(('(?<=\s|\()', '(?=\s|\()')),
              NOR.join(('(?<=\s|\()', '(?=\s|\()')),
              AND.join(('(?<=\s|\))', '(?=\s|\()')),
              NUM.join((LEADING_SPACE, TRAILING_SPACE)),
              LIKE.join((LEADING_SPACE, TRAILING_SPACE)),
              FROM.join((LEADING_SPACE, TRAILING_SPACE)),
              ALPHA.join((LEADING_SPACE, TRAILING_SPACE)),
              ABOVE.join((LEADING_SPACE, TRAILING_SPACE)),
              AFTER.join((LEADING_SPACE, TRAILING_SPACE)),
              BELOW.join((LEADING_SPACE, TRAILING_SPACE)),
              BEFORE.join((LEADING_SPACE, TRAILING_SPACE)),
              PRESENT.join((LEADING_SPACE, '(?=\s|\)|\Z)')),
              STRING,
              )).join(('(', ')')),
    flags=re.IGNORECASE|re.DOTALL)

KEYWORDS = frozenset((LEFT_PARENTHESIS,
                      RIGHT_PARENTHESIS,
                      OR,
                      TO,
                      IS,
                      EQ,
                      NE,
                      LT,
                      LE,
                      GT,
                      GE,
                      NOT,
                      NOR,
                      AND,
                      NUM,
                      LIKE,
                      FROM,
                      ALPHA,
                      ABOVE,
                      AFTER,
                      BELOW,
                      BEFORE,
                      PRESENT,
                      ))
STRUCTURE_TOKENS = frozenset((LEFT_PARENTHESIS,
                              RIGHT_PARENTHESIS,
                              None,
                              ))


class WhereError(Exception):
    pass


class Where(object):

    """"""

    def __init__(self, statement):
        """"""
        self.statement = statement
        self.node = None
        self.tokens = None
        self.token_dispatch = {
            LEFT_PARENTHESIS: self.left_parenthesis,
            RIGHT_PARENTHESIS: self.right_parenthesis,
            OR: self.boolean,
            TO: self.range_end,
            IS: self.is_,
            EQ: self.compare,
            NE: self.compare,
            LT: self.compare,
            LE: self.compare,
            GT: self.compare,
            GE: self.compare,
            NOT: self.invert,
            NOR: self.boolean,
            AND: self.boolean,
            NUM: self.alphanumeric,
            LIKE: self.like,
            FROM: self.range_start,
            ALPHA: self.alphanumeric,
            ABOVE: self.range_start,
            AFTER: self.compare,
            BELOW: self.range_end,
            BEFORE: self.compare,
            PRESENT: self.present,
            None: self.finish,
            False: self.finish,
            }
        self.present_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._field_present,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self.report,
            self.boolean: self._boolean_present,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self._invert_present,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self.report,
            }
        self.range_start_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._field_range_start,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self.report,
            self.boolean: self._boolean_range_start,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self._invert_range_start,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self._alphanumeric_range_start,
            }
        self.range_end_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._value_range_end,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self.report,
            self.boolean: self.report,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self.report,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self.report,
            }
        self.boolean_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._value_boolean,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self._right_parenthesis_boolean,
            self.boolean: self.report,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self.report,
            self.range_start: self.report,
            self.present: self._present_boolean,
            self.alphanumeric: self.report,
            }
        self.invert_dispatch = {
            None: self.report,
            False: self._first_token_invert,
            self.value: self._field_invert,
            self.left_parenthesis: self._left_parenthesis_invert,
            self.right_parenthesis: self.report,
            self.boolean: self._boolean_invert,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self._is_invert,
            self.like: self.report,
            self.invert: self.report,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self.report,
            }
        self.compare_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._field_compare,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self.report,
            self.boolean: self._boolean_compare,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self._invert_compare,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self._alphanumeric_compare,
            }
        self.is_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._field_is,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self.report,
            self.boolean: self._boolean_is,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self.report,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self.report,
            }
        self.like_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._field_like,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self.report,
            self.boolean: self._boolean_like,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self._invert_like,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self.report,
            }
        self.left_parenthesis_dispatch = {
            None: self.report,
            False: self._first_token_left_parenthesis,
            self.value: self.report,
            self.left_parenthesis: self._left_parenthesis_left_parenthesis,
            self.right_parenthesis: self.report,
            self.boolean: self._boolean_left_parenthesis,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self.report,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self.report,
            }
        self.right_parenthesis_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._value_right_parenthesis,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self._right_parenthesis_right_parenthesis,
            self.boolean: self.report,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self.report,
            self.range_start: self.report,
            self.present: self._present_right_parenthesis,
            self.alphanumeric: self.report,
            }
        self.value_dispatch = {
            None: self.report,
            False: self._first_token_field,
            self.value: self.report,
            self.left_parenthesis: self._left_parenthesis_field,
            self.right_parenthesis: self.report,
            self.boolean: self._value_or_field_after_boolean,
            self.range_end: self._range_end_value,
            self.compare: self._compare_value,
            self.is_: self._is_value,
            self.like: self._like_value,
            self.invert: self._value_or_field_after_invert,
            self.range_start: self._range_start_value,
            self.present: self.report,
            self.alphanumeric: self.report,
            }
        self.alphanumeric_dispatch = {
            None: self.report,
            False: self.report,
            self.value: self._field_alphanumeric,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self.report,
            self.boolean: self._boolean_alphanumeric,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self._invert_alphanumeric,
            self.range_start: self.report,
            self.present: self.report,
            self.alphanumeric: self.report,
            }
        self.none_dispatch = {
            None: self.report,
            False: self._finish,
            self.value: self._finish,
            self.left_parenthesis: self.report,
            self.right_parenthesis: self._finish,
            self.boolean: self.report,
            self.range_end: self.report,
            self.compare: self.report,
            self.is_: self.report,
            self.like: self.report,
            self.invert: self.report,
            self.range_start: self.report,
            self.present: self._finish,
            self.alphanumeric: self.report,
            }
        self._processors = None
        self._f_or_v = None
        self._not = None
        
    def lex(self):
        """"""
        tokens = []
        strings = []
        for w in WHERE_RE.split(self.statement):
            if w.lower() in KEYWORDS:
                if strings:
                    tokens.append(''.join([_trim(s) for s in strings if s]))
                    strings.clear()
                tokens.append(w.lower())
            elif w.strip():
                strings.append(w.strip())
        if strings:
            tokens.append(''.join([_trim(s) for s in strings if s]))
            strings.clear()
        tokens.append(None)
        self.tokens = tokens

    def parse(self):
        """"""
        self.node = WhereClause()
        state = False
        for t in self.tokens:
            p = self.token_dispatch.get(t, self.value)
            p(state, t)
            state = p
        if self.node.up is not None:
            self.node = self.node.up

    def validate(self, db, dbset):
        """"""
        if self.node is None:
            return None
        clauses = self.node.get_root().get_clauses_from_root_in_walk_order()

        # Valid nodes, considering (field, condition, value), are like
        # (str, str, str) or (None, None, None).  Mis-spelling a condition
        # will produce the field name ' '.join((field, condition, value, ...))
        # but mis-spelling an operator will produce the value ' '.join((value,
        # operator, field)) and ignore some tokens.
        symbols = []
        nodetokens = []
        fields = set()
        for c in clauses:
            if c.operator is not None:
                nodetokens.append(c.operator)
            expression = []
            if c.field is not None:
                expression.append(c.field)
                if not db.exists(dbset, c.field):
                    fields.add(c.field)
            for s in c.condition, c.value:
                if s is not None:
                    expression.append(s)
            nodetokens.extend(expression)
            if len(expression) in {0, 3}:
                continue
            symbols.append(c.field)

        # Error if len(symbols) not 0.
        if len(symbols):
            return symbols
        
        tokens = [t for t in self.tokens if t not in STRUCTURE_TOKENS]

        # field attribute of each clause must be None or exist as a field.
        if len(fields):
            return fields

        # Error if any tokens have not been kept.
        if len(nodetokens) > len(tokens):
            nodetokens, tokens = tokens, nodetokens
        for t in nodetokens:
            try:
                tokens.remove(t)
            except ValueError:
                pass
        if len(tokens):
            return tokens

        return []

    def evaluate(self, processors):
        """"""
        if self.node is None:
            return None
        self._processors = processors
        try:
            rn = self.node.get_root()
            rn.order_condition_node(self.order_rules_engine)
            rn.evaluate_index_condition_node(self._index_rules_engine)
            if rn.result is not None:
                return
            non_index_nodes = []
            rn.get_non_index_condition_node(non_index_nodes)
            for n in non_index_nodes:
                n.result = WhereResult()
            rn.constraint = WhereConstraint()
            rn.constraint.result = WhereResult()
            rn.constraint.result.answer = processors.get_existence()
            rn.set_non_index_node_constraint(processors.initialize_answer)
            cn = WhereConstraint()
            cn.result = WhereResult()
            processors.initialize_answer(cn)
            for c in {n.constraint for n in non_index_nodes}:
                if c.result:
                    cn.result.answer |= c.result.answer
                else:
                    cn.result.answer = processors.get_existence()
                    break
            for record in processors.get_record(cn.result.answer):
                for n in non_index_nodes:
                    processors.non_index_condition(n, *record)
            self.node.get_root().evaluate_node_result(self._result_rules_engine)
        finally:
            self._processors = None

    def _index_rules_engine(self, op, obj):
        """"""
        if op in {IS,
                  LIKE,
                  PRESENT,
                  EQ,
                  NE,
                  GT,
                  LT,
                  LE,
                  GE,
                  BEFORE,
                  AFTER,
                  (FROM, TO),
                  (FROM, BELOW),
                  (ABOVE, TO),
                  (ABOVE, BELOW),
                  }:
            obj.result = WhereResult()
            self._processors.condition(obj)
        elif op in {NOR, AND, OR}:
            self._processors.operator(obj)
        else:
            self._processors.answer(obj)

    def _result_rules_engine(self, op, obj):
        """"""
        if op in {EQ,
                  GT,
                  LT,
                  LE,
                  GE,
                  BEFORE,
                  AFTER,
                  (FROM, TO),
                  (FROM, BELOW),
                  (ABOVE, TO),
                  (ABOVE, BELOW),
                  }:
            pass
        elif op in {LIKE,
                    PRESENT,
                    NE,
                    }:
            self._processors.not_condition(obj)
        elif op == IS:
            if obj.not_value:
                self._processors.not_condition(obj)
        elif op in {NOR, AND, OR}:
            if obj.result is not obj.left.result:
                self._processors.operator(obj)
        else:
            if obj.result is not obj.up.result:
                self._processors.answer(obj)

    def order_rules_engine(self, obj):
        """Make rightmost leaf node in AND NOR sequences non-index if any.

        Now the OR which breaks the sequence can decide whether to continue
        to combine conditions based on it's left neighbour.

        """
        non_index_node = None
        while True:
            if obj.right is None:
                if non_index_node:
                    if obj.operator == OR:
                        if obj.left.operator != OR:
                            obj.left._swap_node_phrases(non_index_node)
                    elif not (obj.condition in {LIKE, PRESENT, NE} or (
                        obj.condition == IS and obj.not_value)):
                        obj._swap_node_phrases(non_index_node)
                break
            if obj.operator == OR:
                left = obj.left
                if non_index_node:
                    if left.operator != OR:
                        if not (left.condition in {LIKE, PRESENT, NE} or (
                            left.condition == IS and left.not_value)):
                            left._swap_node_phrases(non_index_node)
                non_index_node = None
            elif obj.condition in {LIKE, PRESENT, NE} or (
                obj.condition == IS and obj.not_value):
                non_index_node = obj
            obj = obj.right

    def present(self, s, t):
        """"""
        self.present_dispatch.get(s, self.error)(t)

    def range_start(self, s, t):
        """"""
        self.range_start_dispatch.get(s, self.error)(t)

    def range_end(self, s, t):
        """"""
        self.range_end_dispatch.get(s, self.error)(t)

    def boolean(self, s, t):
        """"""
        self.boolean_dispatch.get(s, self.error)(t)

    def invert(self, s, t):
        """"""
        self.invert_dispatch.get(s, self.error)(t)

    def compare(self, s, t):
        """"""
        self.compare_dispatch.get(s, self.error)(t)

    def is_(self, s, t):
        """"""
        self.is_dispatch.get(s, self.error)(t)

    def like(self, s, t):
        """"""
        self.like_dispatch.get(s, self.error)(t)

    def left_parenthesis(self, s, t):
        """"""
        self.left_parenthesis_dispatch.get(s, self.error)(t)

    def right_parenthesis(self, s, t):
        """"""
        self.right_parenthesis_dispatch.get(s, self.error)(t)

    def value(self, s, t):
        """"""
        self.value_dispatch.get(s, self.error)(t)

    def alphanumeric(self, s, t):
        """"""
        self.alphanumeric_dispatch.get(s, self.error)(t)

    def finish(self, s, t):
        """"""
        self.none_dispatch.get(s, self.error)(t)

    def report(self, t):
        """"""
        raise WhereError('Token not allowed in state')

    def error(self, t):
        """"""
        raise WhereError('Unknown state')

    def _first_token_field(self, t):
        """"""
        wc = WhereClause()
        wc.field = t
        wc.up = self.node
        self.node.down = wc
        self.node = wc

    def _field_range_start(self, t):
        """"""
        self._deferred_field()
        self.node.condition = t

    def _boolean_range_start(self, t):
        """"""
        self._copy_pre_range()
        self.node.condition = t

    def _invert_range_start(self, t):
        """"""
        self._deferred_not_condition()
        self.node.condition = t

    def _alphanumeric_range_start(self, t):
        """"""
        self.node.condition = t

    def _range_start_value(self, t):
        """"""
        self.node.value = t

    def _value_range_end(self, t):
        """"""
        self.node.condition = self.node.condition, t

    def _range_end_value(self, t):
        """"""
        self.node.value = self.node.value, t

    def _value_boolean(self, t):
        """"""
        self._deferred_value()
        s = self.node
        wc = WhereClause()
        s.right = wc
        wc.up = s.up
        wc.left = s
        self.node = wc
        wc.operator = t

    def _present_boolean(self, t):
        """"""
        s = self.node
        wc = WhereClause()
        s.right = wc
        wc.up = s.up
        wc.left = s
        self.node = wc
        wc.operator = t

    def _boolean_left_parenthesis(self, t):
        """"""
        self._deferred_not_phrase()
        wc = self.node
        wc.down = WhereClause()
        wc.down.up = wc
        self.node = wc.down

    def _left_parenthesis_field(self, t):
        """"""
        self.node.field = t

    def _field_compare(self, t):
        """"""
        self._deferred_field()
        self.node.condition = t

    def _boolean_compare(self, t):
        """"""
        self._copy_pre_compare()
        self.node.condition = t

    def _invert_compare(self, t):
        """"""
        self._deferred_not_condition()
        self.node.condition = t

    def _alphanumeric_compare(self, t):
        """"""
        self.node.condition = t

    def _field_is(self, t):
        """"""
        self._deferred_field()
        self.node.condition = t

    def _field_like(self, t):
        """"""
        self._deferred_field()
        self.node.condition = t

    def _compare_value(self, t):
        """"""
        self.node.value = t

    def _is_value(self, t):
        """"""
        self.node.value = t

    def _like_value(self, t):
        """"""
        self.node.value = t

    def _value_right_parenthesis(self, t):
        """"""
        if self.node.up is None:
            raise WhereError('No unmatched left-parentheses')
        self.node = self.node.up

    def _present_right_parenthesis(self, t):
        """"""
        if self.node.up is None:
            raise WhereError('No unmatched left-parentheses')
        self.node = self.node.up

    def _field_present(self, t):
        """"""
        self._deferred_field()
        self.node.condition = t

    def _field_alphanumeric(self, t):
        """"""
        self._deferred_field()
        self.node.alpha = t == ALPHA
        self.node.num = t == NUM

    def _first_token_invert(self, t):
        """"""
        wc = WhereClause()
        wc.not_phrase = True
        wc.up = self.node
        self.node.down = wc
        self.node = wc

    def _value_or_field_after_invert(self, t):
        """"""
        if self.node.condition == IS:
            self._deferred_not_value()
            self.node.value = t
        else:
            self._deferred_not_phrase()
            self.node.field = t

    def _boolean_invert(self, t):
        """"""
        self._not = True

    def _field_invert(self, t):
        """"""
        self._deferred_field()
        self.node.not_condition = True

    def _invert_alphanumeric(self, t):
        """"""
        self._deferred_not_condition()
        self.node.alpha = t == ALPHA
        self.node.num = t == NUM

    def _boolean_alphanumeric(self, t):
        """"""
        self._copy_pre_alphanumeric()
        self.node.alpha = t == ALPHA
        self.node.num = t == NUM

    def _invert_present(self, t):
        """"""
        self._deferred_not_condition()
        self.node.condition = t

    def _boolean_present(self, t):
        """"""
        self._copy_pre_present()
        self.node.condition = t

    def _boolean_is(self, t):
        """"""
        self._copy_pre_is()
        self.node.condition = t

    def _invert_like(self, t):
        """"""
        self._deferred_not_condition()
        self.node.condition = t

    def _boolean_like(self, t):
        """"""
        self._copy_pre_like()
        self.node.condition = t

    def _is_invert(self, t):
        """"""
        self.node.not_value = True

    def _value_or_field_after_boolean(self, t):
        """"""
        self._f_or_v = t

    def _first_token_left_parenthesis(self, t):
        """"""
        wc = WhereClause()
        wc.down = WhereClause()
        wc.down.up = wc
        wc.up = self.node
        self.node.down = wc
        self.node = wc.down

    def _right_parenthesis_boolean(self, t):
        """"""
        s = self.node
        wc = WhereClause()
        s.right = wc
        wc.up = s.up
        wc.left = s
        self.node = wc
        wc.operator = t

    def _left_parenthesis_invert(self, t):
        """"""
        self.node.not_phrase = True

    def _left_parenthesis_left_parenthesis(self, t):
        """"""
        wc = self.node
        wc.down = WhereClause()
        wc.down.up = wc
        self.node = wc.down

    def _right_parenthesis_right_parenthesis(self, t):
        """"""
        self.node = self.node.up

    def _finish(self, t):
        """"""
        self._deferred_value()

    def _deferred_field(self):
        """"""
        if self._f_or_v is not None:
            self.node.field = self._f_or_v
            self._f_or_v = None

    def _deferred_not_phrase(self):
        """"""
        if self._not is not None:
            self.node.not_phrase = self._not
            self._not = None

    def _deferred_not_condition(self):
        """"""
        if self._not is not None:
            self._copy_pre_not_condition()
            self.node.not_condition = self._not
            self._not = None

    def _deferred_not_value(self):
        """"""
        if self._not is not None:
            self.node.not_value = self._not
            self._not = None

    def _deferred_value(self):
        """"""
        if self._f_or_v is not None:
            self._copy_pre_value()
            self.node.value = self._f_or_v
            self._f_or_v = None

    def _copy_pre_compare(self):
        """"""
        self._copy_pre_alphanumeric()
        self.node.num = self.node.left.num
        self.node.alpha = self.node.left.alpha

    def _copy_pre_range(self):
        """"""
        self._copy_pre_alphanumeric()
        self.node.num = self.node.left.num
        self.node.alpha = self.node.left.alpha

    def _copy_pre_alphanumeric(self):
        """"""
        self._copy_pre_not_condition()
        self.node.not_condition = self.node.left.not_condition

    def _copy_pre_present(self):
        """"""
        self._copy_pre_not_condition()
        self.node.not_condition = self.node.left.not_condition

    def _copy_pre_is(self):
        """"""
        self._copy_pre_field()
        self.node.field = self.node.left.field

    def _copy_pre_like(self):
        """"""
        self._copy_pre_not_condition()
        self.node.not_condition = self.node.left.not_condition

    def _copy_pre_not_condition(self):
        """"""
        self._copy_pre_field()
        self.node.field = self.node.left.field

    def _copy_pre_field(self):
        """"""
        self.node.not_phrase = self.node.left.not_phrase

    def _copy_pre_value(self):
        """"""
        snl = self.node.left
        if snl.condition == PRESENT:
            raise WhereError('PRESENT phrase followed by value phrase')
        elif snl.condition == (FROM, TO):
            raise WhereError('FROM-TO phrase followed by value phrase')
        elif snl.condition == (FROM, BELOW):
            raise WhereError('FROM-BELOW phrase followed by value phrase')
        elif snl.condition == (ABOVE, TO):
            raise WhereError('ABOVE-TO phrase followed by value phrase')
        elif snl.condition == (ABOVE, BELOW):
            raise WhereError('ABOVE-BELOW phrase followed by value phrase')
        elif snl.condition == LIKE:
            self._copy_pre_like()
        elif snl.condition == IS:
            self._copy_pre_is()
        else:
            self._copy_pre_compare()
        self.node.condition = snl.condition


class WhereClause(object):

    """Phrase and location in Where specification.

    Navigation attributes are: left, right, up, down.
    (left, up) == (None, None): at most one of these, the root.
    down == None: zero or more of these, leaf nodes.
    up == None: the nodes at the top level.
    (left, right) == (None, <WhereClause instance>): leftmost at each level.
    (left, right) == (<WhereClause instance>, None): rightmost at each level.
    One each of leftmost and rightmost at each level, a double-linked list
    between the two.  Each has same <WhereClause instance> as 'up' attribute.
    down is <WhereClause instance>: next level implements a NOT or a (...).
    left == None implies operator == None.
    left != None implies operator != None.
    In other words, prhases at same level are combined by:
    <left WhereClause phrase attributes> operator <self phrase attributes>,
    and:
    self.result = evaluation(self.down WhereClause) if down != None, or:
    self.result = evaluation(self phrase attribute) if down == None.
    """

    def __init__(self):
        """"""
        # Navigation
        self.left = None
        self.right = None
        self.up = None
        self.down = None
        self.operator = None
        # Phrase
        self.field = None
        self.condition = None
        self.value = None
        self.not_phrase = None
        self.not_condition = None
        self.not_value = None
        self.num = None
        self.alpha = None
        # Evaluation
        self.result = None
        self.constraint = None

    def get_root(self):
        """"""
        wc = self
        while True:
            if wc.left is None and wc.up is None:
                return wc
            if wc.up is not None:
                wc = wc.up
            else:
                wc = wc.left

    def get_clauses_from_current_in_walk_order(self, clauses=None):
        """"""
        if clauses is None:
            clauses = []
        clauses.append(self)
        if self.down is not None:
            self.down.get_clauses_from_current_in_walk_order(clauses=clauses)
        if self.right is not None:
            self.right.get_clauses_from_current_in_walk_order(clauses=clauses)

    def get_clauses_from_root_in_walk_order(self):
        """"""
        clauses = []
        self.get_root().get_clauses_from_current_in_walk_order(clauses=clauses)
        return clauses

    def get_condition(self):
        """"""
        return id(self)

    def evaluate_node_result(self, rules):
        """"""
        if self.down is not None:
            self.down.evaluate_node_result(rules)
            for operator in NOR, AND, OR,:
                n = self.down
                while n:
                    if n.operator == operator:
                        rules(operator, n)
                    n = n.right
            n = self.down
            while n.right:
                n = n.right
            rules(None, n)
        if self.condition is not None:
            rules(self.condition, self)
        if self.right is not None:
            self.right.evaluate_node_result(rules)

    def order_condition_node(self, rules):
        """Make rightmost leaf node in AND NOR sequences non-index if any"""
        if self.down is not None:
            self.down.order_condition_node(rules)
        elif self.right is not None:
            self.right.order_condition_node(rules)
        elif self.up:
            rules(self.up.down)

    def evaluate_index_condition_node(self, rules):
        """"""
        if self.down is not None:
            self.down.evaluate_index_condition_node(rules)
            for operator in NOR, AND, OR,:
                n = self.down
                while n:
                    if n.operator == operator:
                        if n.result and n.left.result:
                            rules(operator, n)
                    n = n.right
            n = self.down
            while True:
                if n.result is None:
                    break
                if n.right is None:
                    rules(None, n)
                    break
                n = n.right
        if self.condition == IS:
            if not self.not_value:
                rules(self.condition, self)
        elif self.condition not in (None, LIKE, PRESENT, NE,):
            rules(self.condition, self)
        if self.operator in {NOR, AND}:
            self.constraint = self.left.constraint
        else:
            self.constraint = WhereConstraint()
        if self.right is not None:
            self.right.evaluate_index_condition_node(rules)

    def _swap_node_phrases(self, left):
        """"""
        for obj in self, left:
            if obj.operator == NOR:
                obj.operator = AND
                obj.not_phrase = not obj.not_phrase
        left.field, self.field = self.field, left.field
        left.condition, self.condition = self.condition, left.condition
        left.value, self.value = self.value, left.value
        left.not_phrase, self.not_phrase = self.not_phrase, left.not_phrase
        left.not_condition, self.not_condition = (
            self.not_condition, left.not_condition)
        left.not_value, self.not_value = self.not_value, left.not_value
        left.num, self.num = self.num, left.num
        left.alpha, self.alpha = self.alpha, left.alpha
    
    def get_non_index_condition_node(self, non_index_nodes):
        """"""
        if self.down is not None:
            self.down.get_non_index_condition_node(non_index_nodes)
        if self.condition == IS:
            if self.not_value:
                non_index_nodes.append(self)
        elif self.condition in (LIKE, PRESENT, NE,):
            non_index_nodes.append(self)
        if self.right is not None:
            self.right.get_non_index_condition_node(non_index_nodes)

    def set_non_index_node_constraint(self, initialize_answer):
        """"""
        if self.left:
            if self.result:
                if self.result.answer is not None:
                    if self.constraint is self.left.constraint:
                        if self.constraint.result:
                            self.constraint.result.answer &= self.result.answer
                        else:
                            self.constraint.result = self.result
                    else:
                        self.constraint.result = self.result
                else:
                    initialize_answer(self)
        elif self.result:
            if self.result.answer is not None:
                self.constraint.result = self.result
            else:
                initialize_answer(self)
        if self.down is not None:
            self.down.set_non_index_node_constraint(initialize_answer)
        if self.right is not None:
            self.right.set_non_index_node_constraint(initialize_answer)
        if self.constraint.result is None:
            if self.up:
                self.constraint = self.up.constraint


class WhereResult(object):
    """"""

    def __init__(self):
        """"""
        self.answer = None


class WhereConstraint(object):
    """"""

    def __init__(self):
        """"""
        self.result = None


def _trim(s):
    if s[0] in '\'"':
        return s[1:-1]
    return s
