# dbapi_test.py
# Copyright 2012 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""dbapi tests"""

import unittest
from copy import copy, deepcopy
import os
import shutil
import subprocess
import ast

from bsddb3.db import (
    DB_CREATE,
    DB_INIT_MPOOL,
    DB_INIT_LOCK,
    DB_INIT_LOG,
    DB_INIT_TXN,
    DB, DB_RDONLY,
    DBEnv,
    DB_RECNO,
    DB_BTREE,
    )

from ..api.test._filespec import (
    samplefilespec,
    SampleValue,
    SampleRecord,
    GAMES_FILE_DEF,
    GAME_FIELD_DEF,
    EVENT_FIELD_DEF,
    SITE_FIELD_DEF,
    DATE_FIELD_DEF,
    ROUND_FIELD_DEF,
    WHITE_FIELD_DEF,
    BLACK_FIELD_DEF,
    RESULT_FIELD_DEF,
    SEVEN_TAG_ROSTER,
    )
from .. import dbapi
from ..api import database, record, recordset, constants


def _api(folder):
    class P:
        def __init__(self, *a, **b):
            pass
    class S(P):
        pass
    return dbapi._DBapi(
        P,
        S,
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'dbapi_tests', folder)),
        None)


def api(folder):
    return dbapi.DBapi(
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'dbapi_tests', folder)))


def bitapi(folder):
    return dbapi.DBbitapi(
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'dbapi_tests', folder)),
        {'flags': (DB_CREATE |
                   DB_INIT_MPOOL |
                   DB_INIT_LOCK |
                   DB_INIT_LOG |
                   DB_INIT_TXN)})


def _delete_folder(folder):
    shutil.rmtree(
        os.path.expanduser(os.path.join('~', 'dbapi_tests', folder)),
        ignore_errors=True)


class SampleRecordAssumptions(unittest.TestCase):

    def setUp(self):
        self.sr = SampleRecord()
        self.sv = SampleValue()

    def tearDown(self):
        self.sr = None
        self.sv = None

    def test____assumptions_01(self):
        self.assertEqual(len(self.sr.__dict__), 8)
        self.assertIsInstance(self.sr.key, record.KeyData)
        self.assertIsInstance(self.sr.value, record.Value)
        self.assertIsInstance(self.sr.value, SampleValue)
        self.assertEqual(self.sr.record, None)
        self.assertEqual(self.sr.database, None)
        self.assertEqual(self.sr.dbname, None)
        self.assertEqual(self.sr.srkey, None)
        self.assertEqual(self.sr.srvalue, None)
        self.assertEqual(self.sr.srindex, None)
        self.assertEqual(len(self.sv.__dict__), 0)
        self.assertIsInstance(self.sv, record.Value)
        self.assertIsInstance(self.sv, SampleValue)

    def test____assumptions_02(self):
        self.assertRaisesRegex(
            ValueError,
            "malformed node or string: \[\]",
            self.sr.load_record,
            *((0, []),))
        self.assertRaisesRegex(
            TypeError,
            "__dict__ must be set to a dictionary, not a 'list'",
            self.sr.load_record,
            *((0, '[]'),))
        self.sr.load_record((0, '{}'))
        self.assertEqual(len(self.sr.key.__dict__), 1)
        self.assertEqual(self.sr.key.recno, 0)
        self.assertEqual(len(self.sr.value.__dict__), 0)

    def test____assumptions_03(self):
        self.sr.load_record((0, repr({})))
        self.assertEqual(len(self.sr.value.__dict__), 0)

    def test____assumptions_04(self):
        # This happens, but will not work in database.
        self.sr.load_record(('0', '{}'))
        self.assertEqual(self.sr.key.recno, b'0')

    def test____assumptions_05(self):
        # This happens, but will not work in database.
        self.sr.load_record(([], '{}'))
        self.assertEqual(self.sr.key.recno, [])

    def test____assumptions_06(self):
        self.assertEqual(SEVEN_TAG_ROSTER,
                         (EVENT_FIELD_DEF,
                          SITE_FIELD_DEF,
                          DATE_FIELD_DEF,
                          ROUND_FIELD_DEF,
                          WHITE_FIELD_DEF,
                          BLACK_FIELD_DEF,
                          RESULT_FIELD_DEF,
                          ))
        value = {GAME_FIELD_DEF:'game score',
                 WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 EVENT_FIELD_DEF:'Match',
                 ROUND_FIELD_DEF:'1',
                 DATE_FIELD_DEF:'today',
                 RESULT_FIELD_DEF:'1-0',
                 SITE_FIELD_DEF:'club venue',
                 'attribute':None,
                 }
        self.sr.load_record((0, repr(value)))
        self.assertEqual(self.sr.value.__dict__, value)
        pv = self.sr.value.pack()
        # pv[0] is not necessarely equal to repr(value)
        self.assertEqual(ast.literal_eval(pv[0]),
                         ast.literal_eval(repr(value)))
        self.assertEqual(pv[1],
                         {WHITE_FIELD_DEF:['A Jones'],
                          BLACK_FIELD_DEF:['B Smith'],
                          EVENT_FIELD_DEF:['Match'],
                          ROUND_FIELD_DEF:['1'],
                          DATE_FIELD_DEF:['today'],
                          RESULT_FIELD_DEF:['1-0'],
                          SITE_FIELD_DEF:['club venue'],
                          })

    def test____assumptions_07(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.sr.load_record((0, repr(value)))
        pv = self.sr.value.pack()
        self.assertEqual(pv[1],
                         {WHITE_FIELD_DEF:['A Jones'],
                          BLACK_FIELD_DEF:['B Smith'],
                          })


class _DatabaseEncoders(unittest.TestCase):

    def setUp(self):
        self.sde = dbapi._DatabaseEncoders()

    def tearDown(self):
        pass

    def test____assumptions(self):
        self.assertEqual(repr(57), '57')

    def test_encode_record_number_01(self):
        self.assertEqual(self.sde.encode_record_number(57), b'57')

    def test_decode_record_number_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'int' object has no attribute 'decode'",
            self.sde.decode_record_number,
            57)
        self.assertEqual(self.sde.decode_record_number(b'57'), 57)

    def test_encode_record_selector(self):
        self.assertEqual(self.sde.encode_record_selector('57'), b'57')


class _DBapi_init(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        self.assertEqual(
            dbapi._DBapi(None, None, {}, '', {}).__dict__,
            dict(_main={},
                 _DBenvironment={},
                 _home='',
                 _dbenv=None,
                 _dbtxn=None,
                 _associate={}))
        self.assertIs(
            dbapi._DBapi(None, None, {}, '', {}).engine_uses_bytes_or_str,
            dbapi._DBapi.engine_uses_bytes_or_str)
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 5 required positional arguments: ",
                "'primary_class', 'secondary_class', 'DBnames', 'DBhome', ",
                "and 'DBenvironment'",
                )),
            dbapi._DBapi,
            )
        self.assertRaisesRegex(
            AttributeError,
            "".join((
                "'NoneType' object has no attribute 'rfind'",
                )),
            dbapi._DBapi,
            *(None, None, None, None, None))
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "'NoneType' object is not iterable",
                )),
            dbapi._DBapi,
            *(None, None, None, '', None))
        self.assertIsInstance(
            dbapi._DBapi(None, None, {}, '', None),
            dbapi._DBapi)
        self.assertRaisesRegex(
            AttributeError,
            "".join((
                "'NoneType' object has no attribute 'setdefault'",
                )),
            dbapi._DBapi,
            *(None, None, dict(t=None), '', None))
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "'NoneType' object is not callable",
                )),
            dbapi._DBapi,
            *(None, None, dict(t={}), '', None))
        class C:
            def __init__(self, *a, **b):
                pass
        self.assertIsInstance(
            dbapi._DBapi(C, None, dict(t={}), '', None),
            dbapi._DBapi)
        self.assertIsInstance(
            dbapi._DBapi(
                C,
                None,
                dict(t=dict(primary='key')),
                '',
                None),
            dbapi._DBapi)
        self.assertRaisesRegex(
            AttributeError,
            "".join((
                "'dict' object has no attribute 'field_name'",
                )),
            dbapi._DBapi,
            *(C,
              None,
              dict(t=dict(primary='key',
                          secondary={'key':None})),
              '',
              None))
        self.assertRaisesRegex(
            dbapi.DBapiError,
            "".join((
                "DB name key requested for secondary name key in primary ",
                "name t is already specified",
                )),
            dbapi._DBapi,
            *(C,
              None,
              dict(t=dict(primary='key',
                          secondary={'key':'key'})),
              '',
              None))
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "'NoneType' object is not callable",
                )),
            dbapi._DBapi,
            *(C,
              None,
              dict(t=dict(primary='key',
                          secondary={'lock':'lock'})),
              '',
              None))
        self.assertIsInstance(
            dbapi._DBapi(C,
                         C,
                         dict(t=dict(primary='key',
                                     secondary={'lock':'lock'})),
                         '',
                         None),
            dbapi._DBapi)
        self.assertIsInstance(
            dbapi._DBapi(C,
                         C,
                         dict(t=dict(primary='key',
                                     secondary={'lock':'locks'})),
                         '',
                         None),
            dbapi._DBapi)

    def test___init___02(self):
        class C:
            def __init__(self, *a, **b):
                pass
        api = dbapi._DBapi(C,
                           C,
                           dict(t=dict(primary='key',
                                       secondary={'lock':'locks'})),
                           '',
                           None)
        self.assertEqual(api._associate, dict(t=dict(lock='locks', t='key')))
        self.assertEqual(len(api._main), 2)
        self.assertIsInstance(api._main['key'], C)
        self.assertIsInstance(api._main['locks'], C)
        self.assertEqual(len(api.__dict__), 6)


class _DBapi_backout(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
        api = dbapi._DBapi(C,
                           C,
                           dict(t=dict(primary='key',
                                       secondary={'lock':'locks'})),
                           '',
                           None)
        self.assertEqual(api.backout(), None)


class _DBapi_close_context(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
            def close(self):
                pass
        api = dbapi._DBapi(C,
                           C,
                           dict(t=dict(primary='key',
                                       secondary={'lock':'locks'})),
                           '',
                           None)
        self.assertEqual(api.close_context(), None)


class _DBapi_close_contexts(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
        api = dbapi._DBapi(C,
                           C,
                           dict(t=dict(primary='key',
                                       secondary={'lock':'locks'})),
                           '',
                           None)
        self.assertEqual(api.close_contexts(None), None)


class _DBapi_close_database(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
            def close(self):
                pass
        api = dbapi._DBapi(C,
                           C,
                           dict(t=dict(primary='key',
                                       secondary={'lock':'locks'})),
                           '',
                           None)
        self.assertEqual(api.close_database(), None)


class _DBapi_commit(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
        api = dbapi._DBapi(C,
                           C,
                           dict(t=dict(primary='key',
                                       secondary={'lock':'locks'})),
                           '',
                           None)
        self.assertEqual(api.commit(), None)


class _DBapi_db_compatibility_hack(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
        api = dbapi._DBapi(C,
                           C,
                           dict(t=dict(primary='key',
                                       secondary={'lock':'locks'})),
                           '',
                           None)
        self.assertEqual(api.db_compatibility_hack('value', 'key'), 'value')


class _DBapi_do_deferred_updates(unittest.TestCase):

    def setUp(self):
        self.api = _api('do_deferred_updates')

    def tearDown(self):
        _delete_folder('do_deferred_updates')

    def test_do_deferred_updates_01(self):
        self.assertRaisesRegex(
            dbapi.DBapiError,
            "'pyscript' is not an existing file",
            self.api.do_deferred_updates,
            *('pyscript', 'filepath'))
        self.assertRaisesRegex(
            dbapi.DBapiError,
            ''.join(("'/home/roger/dbapi_tests/do_deferred_updates' is ",
                     "not an existing file",
                     )),
            self.api.do_deferred_updates,
            *(self.api._home, 'filepath'))
        self.assertRaisesRegex(
            dbapi.DBapiError,
            ''.join(("'/home/roger/dbapi_tests/do_deferred_updates' is ",
                     "not an existing file",
                     )),
            self.api.do_deferred_updates,
            *(self.api._home, (self.api._home, 'filepath')))
        script = os.path.join(self.api._home, 'script')
        os.makedirs(self.api._home)
        open(script, 'w').close()
        sp = self.api.do_deferred_updates(script, self.api._home)
        self.assertIsInstance(sp, subprocess.Popen)
        sp.wait()


class _DBapi_exists(unittest.TestCase):

    def setUp(self):
        self.api = _api('exists')

    def tearDown(self):
        _delete_folder('exists')

    def test_exists_01(self):
        self.assertEqual(self.api.exists(None, None), False)
        self.assertEqual(self.api.exists('games', None), False)
        self.assertEqual(self.api.exists('games', 'Site'), True)


class _DBapi_files_exist(unittest.TestCase):

    def setUp(self):
        self.api = _api('files_exist')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}
        for c in self.api._main.values():
            c.open_root = lambda a : None
            c.close = lambda : None

    def tearDown(self):
        _delete_folder('files_exist')

    def test_exists_01(self):
        os.makedirs(self.api._home)
        self.assertEqual(self.api.files_exist(), None)
        self.api.open_context()
        #self.assertEqual(self.api.files_exist(), True)
        self.api.close_database()
        #self.assertEqual(self.api.files_exist(), True)


class _DBapi_database_cursor(unittest.TestCase):

    def setUp(self):
        class C:
            pass
        self.c = C
        self.api = _api('database_cursor')
        for c in self.api._main.values():
            c.make_cursor = lambda a, b : C()

    def tearDown(self):
        _delete_folder('database_cursor')

    def test_database_cursor_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.database_cursor,
            *(None, None))
        self.assertIsInstance(
            self.api.database_cursor('games', 'Site'),
            self.c)
        self.assertIsInstance(
            self.api.database_cursor('games', 'games'),
            self.c)


class _DBapi_repair_cursor(unittest.TestCase):

    def setUp(self):
        self.api = _api('database_cursor')

    def tearDown(self):
        _delete_folder('database_cursor')

    def test_repair_cursor_01(self):
        class C:
            pass
        c = C()
        self.assertIs(self.api.repair_cursor(c), c)


class _DBapi_get_database_folder(unittest.TestCase):

    def setUp(self):
        self.api = _api('get_database_folder')

    def tearDown(self):
        _delete_folder('get_database_folder')

    def test_get_database_folder_01(self):
        self.assertEqual(
            self.api.get_database_folder(),
            os.path.expanduser(
                os.path.join('~', 'dbapi_tests', 'get_database_folder')))


class _DBapi_get_database(unittest.TestCase):

    def setUp(self):
        class O:
            pass
        self.o = O
        self.api = _api('get_database')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}
        for c in self.api._main.values():
            c._object = None
            c.open_root = lambda a : None

    def tearDown(self):
        _delete_folder('get_database')

    def test_get_database_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_database,
            *(None, None))
        self.assertEqual(self.api.get_database('games', 'Site'), None)
        self.assertEqual(self.api.get_database('games', 'games'), None)
        self.api.open_context()
        for c in self.api._main.values():
            c._object = self.o()
        self.assertIsInstance(
            self.api.get_database('games', 'Site'),
            self.o)
        self.assertIsInstance(
            self.api.get_database('games', 'games'),
            self.o)


class _DBapi_get_database_instance(unittest.TestCase):

    def setUp(self):
        class C:
            pass
        self.c = C
        self.api = _api('get_database_instance')
        for c in self.api._main:
            self.api._main[c] = C()

    def tearDown(self):
        _delete_folder('get_database_instance')

    def test_get_database_instance_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_database_instance,
            *(None, None))
        self.assertIsInstance(
            self.api.get_database_instance('games', 'Site'),
            self.c)
        self.assertIsInstance(
            self.api.get_database_instance('games', 'games'),
            self.c)


class _DBapi_get_first_primary_key_for_index_key(unittest.TestCase):
    # Test results are suspicious but may be correct.

    def setUp(self):
        def f(a):
            raise dbapi.DBapiError(
                "get_first_primary_key_for_index_key for primary table")
        self.api = _api('get_first_primary_key_for_index_key')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}
        for v in self.api._main.values():
            v.get_first_primary_key_for_index_key = f
            v.open_root = lambda a : None

    def tearDown(self):
        _delete_folder('get_first_primary_key_for_index_key')

    def test_get_first_primary_key_for_index_key_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_first_primary_key_for_index_key,
            *(None, None, None))
        #self.assertRaisesRegex(
        #    AttributeError,
        #    ''.join((
        #        "'NoneType' object has no attribute ",
        #        "'execute'")),
        #    self.api.get_first_primary_key_for_index_key,
        #    *('games', 'Site', 'k'))
        self.assertRaisesRegex(
            dbapi.DBapiError,
            "get_first_primary_key_for_index_key for primary table",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'games', 'k'))
        self.api.open_context()
        #self.assertEqual(
        #    self.api.get_first_primary_key_for_index_key('games', 'games', 'k'),
        #    None)
        self.assertRaisesRegex(
            dbapi.DBapiError,
            "get_first_primary_key_for_index_key for primary table",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'games', 'k'))


class _DBapi_get_primary_record(unittest.TestCase):

    def setUp(self):
        self.api = _api('get_primary_record')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}
        for v in self.api._main.values():
            v.get_primary_record = lambda key : key
            v.open_root = lambda a : None

    def tearDown(self):
        _delete_folder('get_primary_record')

    def test_get_primary_record_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_primary_record,
            *(None, None))
        self.api.open_context()
        self.assertEqual(self.api.get_primary_record('games', 'k'), 'k')


class _DBapi_is_primary(unittest.TestCase):

    def setUp(self):
        self.api = _api('is_primary')
        for v in self.api._main.values():
            v.is_primary = lambda : 'is_primary'

    def tearDown(self):
        _delete_folder('is_primary')

    def test_is_primary_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.is_primary,
            *(None, None))
        self.assertEqual(self.api.is_primary('games', 'games'), 'is_primary')
        self.assertEqual(self.api.is_primary('games', 'Site'), 'is_primary')


class _DBapi_is_primary_recno(unittest.TestCase):

    def setUp(self):
        self.api = _api('is_primary_recno')

    def tearDown(self):
        _delete_folder('is_primary_recno')

    def test_is_primary_recno_01(self):
        self.assertEqual(
            self.api.is_primary_recno('Berkeley DB compatibility'), True)


class _DBapi_is_recno(unittest.TestCase):

    def setUp(self):
        self.api = _api('is_recno')
        for v in self.api._main.values():
            v.is_recno = lambda : 'is_recno'

    def tearDown(self):
        _delete_folder('is_recno')

    def test_is_recno_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.is_recno,
            *(None, None))
        self.assertEqual(self.api.is_recno('games', 'games'), 'is_recno')
        self.assertEqual(self.api.is_recno('games', 'Site'), 'is_recno')


class _DBapi_open_contexts(unittest.TestCase):

    def setUp(self):
        self.api = _api('open_contexts')
        for v in self.api._main.values():
            v.open_contexts = lambda a : None

    def tearDown(self):
        _delete_folder('open_contexts')

    def test_open_contexts_01(self):
        self.assertEqual(self.api.open_contexts('DPT compatibility'), None)


class _DBapi_allocate_and_open_contexts(unittest.TestCase):

    def setUp(self):
        self.api = _api('allocate_and_open_contexts')
        for v in self.api._main.values():
            v.allocate_and_open_contexts = lambda a : None

    def tearDown(self):
        _delete_folder('allocate_and_open_contexts')

    def test_allocate_and_open_contexts_01(self):
        self.assertEqual(self.api.allocate_and_open_contexts(
            'DPT compatibility'), None)


class _DBapi_get_packed_key(unittest.TestCase):

    def setUp(self):
        self.api = _api('get_packed_key')

    def tearDown(self):
        _delete_folder('get_packed_key')

    def test_get_packed_key_01(self):
        instance = record.Record(record.Key, record.Value)
        self.assertEqual(
            self.api.get_packed_key('Berkeley DB compatibility',
                                    instance).decode(),
            repr(instance.key.pack()))


class _DBapi_decode_as_primary_key(unittest.TestCase):

    def setUp(self):
        self.api = _api('decode_as_primary_key')

    def tearDown(self):
        _delete_folder('decode_as_primary_key')

    def test_decode_as_primary_key_01(self):
        #self.assertEqual(self.api.decode_as_primary_key(
        #    'Ignored in sqlite3', 10), 10)
        srkey = self.api.encode_record_number(3456)
        self.assertEqual(self.api.decode_as_primary_key(
            'Ignored in sqlite3', srkey), 3456)


class _DBapi_encode_primary_key(unittest.TestCase):

    def setUp(self):
        self.api = _api('encode_primary_key')

    def tearDown(self):
        _delete_folder('encode_primary_key')

    def test_encode_primary_key_01(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.load(23)
        self.assertEqual(
            self.api.encode_primary_key('games', instance),
            b'23')


class _DBapi_set_defer_update(unittest.TestCase):

    def setUp(self):
        self.api = _api('set_defer_update')
        for v in self.api._main.values():
            v.close = lambda : None

    def tearDown(self):
        _delete_folder('set_defer_update')

    def test_set_defer_update_01(self):
        self.assertEqual(
            self.api.set_defer_update(duallowed='kkk'),
            'kkk')


class _DBapi_unset_defer_update(unittest.TestCase):

    def setUp(self):
        self.api = _api('unset_defer_update')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}
        for v in self.api._main.values():
            v.open_root = lambda a : None

    def tearDown(self):
        _delete_folder('unset_defer_update')

    def test_unset_defer_update_01(self):
        self.assertEqual(
            self.api.unset_defer_update(),
            True)


class _DBapi_use_deferred_update_process(unittest.TestCase):

    def setUp(self):
        self.api = _api('use_deferred_update_process')

    def tearDown(self):
        _delete_folder('use_deferred_update_process')

    def test_unset_defer_update_01(self):
        self.assertRaisesRegex(
            dbapi.DBapiError,
            "use_deferred_update_process not implemented",
            self.api.use_deferred_update_process,
            **{})


class _DBapi_initial_database_size(unittest.TestCase):

    def setUp(self):
        self.api = _api('initial_database_size')

    def tearDown(self):
        _delete_folder('initial_database_size')

    def test_initial_database_size_01(self):
        self.assertEqual(
            self.api.initial_database_size(),
            True)


class _DBapi_increase_database_size(unittest.TestCase):

    def setUp(self):
        self.api = _api('increase_database_size')

    def tearDown(self):
        _delete_folder('increase_database_size')

    def test_initial_database_size_01(self):
        self.assertEqual(
            self.api.increase_database_size(),
            None)


class _DBapi_do_database_task(unittest.TestCase):

    def setUp(self):
        def f(obj, lw, **tma):
            pass
        self.f = f
        self.api = _api('do_database_task')

    def tearDown(self):
        _delete_folder('do_database_task')

    def test_initial_database_size_01(self):
        self.assertEqual(
            self.api.do_database_task(self.f),
            None)


class _DBapi_make_recordset_key(unittest.TestCase):

    def setUp(self):
        self.api = _api('make_recordset_key')
        for v in self.api._main.values():
            v.populate_recordset_key = lambda a, b : None
            v._object = None
            v._recordsets = {}

    def tearDown(self):
        _delete_folder('make_recordset_key')

    def test_make_recordset_key_01(self):
        self.assertIsInstance(
            self.api.make_recordset_key('games', 'games'),
            recordset.Recordset)


class _DBapi_make_recordset_key_startswith(unittest.TestCase):

    def setUp(self):
        self.api = _api('make_recordset_key_startswith')
        for v in self.api._main.values():
            v.populate_recordset_key_startswith = lambda a, b : None
            v._object = None
            v._recordsets = {}

    def tearDown(self):
        _delete_folder('make_recordset_key_startswith')

    def test_make_recordset_key_startswith_01(self):
        self.assertIsInstance(
            self.api.make_recordset_key_startswith('games', 'games'),
            recordset.Recordset)


class _DBapi_make_recordset_key_range(unittest.TestCase):

    def setUp(self):
        self.api = _api('make_recordset_key_range')
        for v in self.api._main.values():
            v.populate_recordset_key_range = lambda a, b, c : None
            v._object = None
            v._recordsets = {}

    def tearDown(self):
        _delete_folder('make_recordset_key_range')

    def test_make_recordset_key_range_01(self):
        self.assertIsInstance(
            self.api.make_recordset_key_range('games', 'games'),
            recordset.Recordset)


class _DBapi_make_recordset_all(unittest.TestCase):

    def setUp(self):
        self.api = _api('make_recordset_all')
        for v in self.api._main.values():
            v.populate_recordset_all = lambda a : None
            v._object = None
            v._recordsets = {}

    def tearDown(self):
        _delete_folder('make_recordset_all')

    def test_make_recordset_all_01(self):
        self.assertIsInstance(
            self.api.make_recordset_all('games', 'games'),
            recordset.Recordset)


class _DBapi_recordset_for_segment(unittest.TestCase):

    def setUp(self):
        self.api = _api('recordset_for_segment')
        for v in self.api._main.values():
            v.populate_recordset_from_segment = lambda a, b : a
            v._object = None
            v._recordsets = {}

    def tearDown(self):
        _delete_folder('recordset_for_segment')

    def test_recordset_for_segment_01(self):
        self.assertIsInstance(
            self.api.recordset_for_segment(
                recordset.Recordset(dbhome=self.api, dbset='games'),
                'games',
                None),
            recordset.Recordset)


class _DBapi_file_records_under(unittest.TestCase):

    def setUp(self):
        self.api = _api('file_records_under')
        for v in self.api._main.values():
            v.file_records_under = lambda a, b : None
            v._object = None
            v._recordsets = {}

    def tearDown(self):
        _delete_folder('file_records_under')

    def test_file_records_under_01(self):
        napi = _api('file_records_under')
        for v in napi._main.values():
            v.file_records_under = lambda a, b : None
            v._object = None
            v._recordsets = {}
        #self.assertRaisesRegex(
        #    database.DatabaseError,
        #    "Record set was not created from this database instance",
        #    self.api.file_records_under,
        #    *('games',
        #      'Site',
        #      recordset.Recordset(dbhome=napi, dbset='sames'),
        #      None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Record set was not created from dbset database",
            self.api.file_records_under,
            *('games',
              'games',
              recordset.Recordset(dbhome=self.api, dbset='Site'),
              None))
        self.assertEqual(
            self.api.file_records_under(
                'games',
                'games',
                recordset.Recordset(dbhome=self.api, dbset='games'),
                None),
            None)


class _DBapi_is_engine_uses_bytes(unittest.TestCase):

    def setUp(self):
        self.api = _api('is_engine_uses_bytes')

    def tearDown(self):
        _delete_folder('is_engine_uses_bytes')

    def test_is_engine_uses_bytes_01(self):
        self.assertEqual(
            self.api.is_engine_uses_bytes(),
            True)


class _DBapi_is_engine_uses_str(unittest.TestCase):

    def setUp(self):
        self.api = _api('is_engine_uses_str')

    def tearDown(self):
        _delete_folder('is_engine_uses_str')

    def test_is_engine_uses_str_01(self):
        self.assertEqual(
            self.api.is_engine_uses_str(),
            False)


class _DBapi_start_transaction(unittest.TestCase):

    def setUp(self):
        self.api = _api('start_transaction')

    def tearDown(self):
        _delete_folder('start_transaction')

    def test_start_transaction_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'txn_begin'",
            self.api.start_transaction)


class _DBapi_cede_contexts_to_process(unittest.TestCase):

    def setUp(self):
        self.api = _api('cede_contexts_to_process')

    def tearDown(self):
        _delete_folder('cede_contexts_to_process')

    def test_cede_contexts_to_process_01(self):
        self.assertEqual(
            self.api.cede_contexts_to_process({}),
            None)


class DBbitapi_method_classes(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_method_classes_01(self):
        bit = dbapi.DBbitapi
        api = dbapi._DBapi
        base = database.Database
        enc = dbapi._DatabaseEncoders

        self.assertEqual(
            len([a for a in dir(bit)
                 if not a.startswith('__') and not a.endswith('__')]),
            50)
        self.assertEqual(
            len([a for a in dir(api)
                 if not a.startswith('__') and not a.endswith('__')]),
            50)
        self.assertEqual(
            len([a for a in dir(base)
                 if not a.startswith('__') and not a.endswith('__')]),
            25)
        self.assertEqual(
            len([a for a in dir(enc)
                 if not a.startswith('__') and not a.endswith('__')]),
            4)

        # Methods inherited from _DBapi
        self.assertIs(api.backout, bit.backout)
        self.assertIs(api.close_contexts, bit.close_contexts)
        self.assertIs(api.close_database, bit.close_database)
        self.assertIs(api.commit, bit.commit)
        self.assertIs(api.db_compatibility_hack, bit.db_compatibility_hack)
        self.assertIs(api.do_deferred_updates, bit.do_deferred_updates)
        self.assertIs(api.exists, bit.exists)
        self.assertIs(api.files_exist, bit.files_exist)
        self.assertIs(api.database_cursor, bit.database_cursor)
        self.assertIs(api.repair_cursor, bit.repair_cursor)
        self.assertIs(api.get_database_folder, bit.get_database_folder)
        self.assertIs(api.get_database, bit.get_database)
        self.assertIs(api.get_database_instance, bit.get_database_instance)
        self.assertIs(
            api.get_first_primary_key_for_index_key,
            bit.get_first_primary_key_for_index_key)
        self.assertIs(api.get_primary_record, bit.get_primary_record)
        self.assertIs(api.is_primary, bit.is_primary)
        self.assertIs(api.is_primary_recno, bit.is_primary_recno)
        self.assertIs(api.is_recno, bit.is_recno)
        self.assertIs(api.open_contexts, bit.open_contexts)
        self.assertIs(api.allocate_and_open_contexts,
                      bit.allocate_and_open_contexts)
        self.assertIs(api.get_packed_key, bit.get_packed_key)
        self.assertIs(api.decode_as_primary_key, bit.decode_as_primary_key)
        self.assertIs(api.encode_primary_key, bit.encode_primary_key)
        self.assertIs(api.set_defer_update, bit.set_defer_update)
        self.assertIs(api.unset_defer_update, bit.unset_defer_update)
        self.assertIs(
            api.use_deferred_update_process,
            bit.use_deferred_update_process)
        self.assertIs(api.initial_database_size, bit.initial_database_size)
        self.assertIs(api.increase_database_size, bit.increase_database_size)
        self.assertIs(api.do_database_task, bit.do_database_task)
        self.assertIs(api.make_recordset_key, bit.make_recordset_key)
        self.assertIs(
            api.make_recordset_key_startswith,
            bit.make_recordset_key_startswith)
        self.assertIs(
            api.make_recordset_key_range,
            bit.make_recordset_key_range)
        self.assertIs(api.make_recordset_all, bit.make_recordset_all)
        self.assertIs(api.recordset_for_segment, bit.recordset_for_segment)
        self.assertIs(api.file_records_under, bit.file_records_under)
        self.assertIs(api.is_engine_uses_bytes, bit.is_engine_uses_bytes)
        self.assertIs(api.is_engine_uses_str, bit.is_engine_uses_str)
        self.assertIs(api.start_transaction, bit.start_transaction)
        self.assertIs(api.cede_contexts_to_process,
                      bit.cede_contexts_to_process)
        
        # Methods inherited from _DatabaseEncoders
        self.assertIs(enc.encode_record_number, bit.encode_record_number)
        self.assertIs(enc.decode_record_number, bit.decode_record_number)
        self.assertIs(enc.encode_record_selector, bit.encode_record_selector)
        
        # Methods inherited from Database
        self.assertIs(base.deferred_update_housekeeping,
                      bit.deferred_update_housekeeping)

        # _DBapi methods overridden
        self.assertIsNot(api.__init__, bit.__init__)
        self.assertIsNot(api.open_context, bit.open_context)
        self.assertIsNot(api.close_context, bit.close_context)
        self.assertIsNot(api.delete_instance, bit.delete_instance)
        self.assertIsNot(api.edit_instance, bit.edit_instance)
        self.assertIsNot(api.put_instance, bit.put_instance)

        # _DatabaseEncoders methods overridden
        self.assertIsNot(enc.encode_record_key, bit.encode_record_key)


class DBbitapi_init(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 1 required positional argument: ",
                "'DBnames'",
                )),
            dbapi.DBbitapi,
            )
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 2 required positional arguments: ",
                "'DBhome' and 'DBenvironment'",
                )),
            dbapi.DBbitapi,
            *(samplefilespec(),))
        self.assertRaisesRegex(
            AttributeError,
            "".join((
                "'NoneType' object has no attribute 'rfind'",
                )),
            dbapi.DBbitapi,
            *(samplefilespec(),),
            **dict(DBenvironment='', DBhome=None))
        self.assertRaisesRegex(
            AttributeError,
            "".join((
                "'NoneType' object has no attribute 'rfind'",
                )),
            dbapi.DBbitapi,
            *(samplefilespec(), None, ''))

    def test___init___02(self):
        api = dbapi.DBbitapi(
            dict(
                t=dict(
                    primary='key',
                    fields=dict(lock={}, key={}),
                    filedesc=dict(fileorg=None),
                    secondary=dict(key='lock'),
                    )),
            DBhome='',
            DBenvironment=None)
        self.assertEqual(api._home, '')
        self.assertEqual(api._dbenv, None)
        self.assertEqual(api._associate, dict(t=dict(key='lock', t='key')))
        self.assertEqual(len(api._main), 2)
        self.assertIsInstance(
            api._main['key'], dbapi.DBbitPrimary)
        self.assertIsInstance(
            api._main['lock'], dbapi.DBbitSecondary)
        self.assertIsInstance(api._control, dbapi.DBbitControlFile)
        self.assertEqual(len(api.__dict__), 7)
        self.assertIsInstance(
            dbapi.DBbitapi(samplefilespec(), '', None),
            dbapi.DBbitapi)
        self.assertIsInstance(
            dbapi.DBbitapi(DBnames=samplefilespec(),
                           DBhome='',
                           DBenvironment=None),
            dbapi.DBbitapi)


class DBbitapi_encode_record_key(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('encode_record_key')

    def tearDown(self):
        _delete_folder('encode_record_key')

    def test_encode_record_key_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'encode'",
            self.api.encode_record_key,
            *(None,))
        self.assertEqual(self.api.encode_record_key('key'), b'key')
        self.assertEqual(
            self.api.encode_record_key(repr(('d', 'e', 'f'))),
            b"('d', 'e', 'f')")


class DBbitapi_open_context(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('open_context')

    def tearDown(self):
        _delete_folder('open_context')

    def test_open_context_01(self):
        self.assertEqual(self.api._control._control_object, None)
        self.assertEqual(self.api.open_context(), True)
        self.assertIsInstance(self.api._control._control_object, type(DB()))


class DBbitapi_close_context(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('close_context')

    def tearDown(self):
        _delete_folder('close_context')

    def test_close_context_01(self):
        self.assertEqual(self.api._control._control_object, None)
        self.assertEqual(self.api.close_context(), None)
        self.assertEqual(self.api._control._control_object, None)

    def test_close_context_02(self):
        self.assertEqual(self.api._control._control_object, None)
        self.assertEqual(self.api.open_context(), True)
        self.assertEqual(self.api.close_context(), None)
        self.assertEqual(self.api._control._control_object, None)


class DBbitapi_put_instance(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('put_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('put_instance')

    def test_put_instance_05(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=1,
                                    Game__bits=0,
                                    Game__list=0,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1,) is the record number.
                 records=dict(White=(('A Jones', 0, 1, (1,)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__exist=(b''.join((b'\x40', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 1, (1,)),),
                              ),
                 ))

    def test_put_instance_06(self):
        # commit inside record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(2):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
            self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=2,
                                    Game__bits=0,
                                    Game__list=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers whight could be used.
                 # White will get one and Black the other, at random.
                 records=dict(White=(('A Jones', 0, 2, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 3)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 3)])),
                              Game__exist=(b''.join((b'\x60', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 2, (1, 2)),),
                              ),
                 ))

    def test_put_instance_07(self):
        # commit inside record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
            self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=3,
                                    Game__bits=0,
                                    Game__list=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 3, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 4)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 4)])),
                              Game__exist=(b''.join((b'\x70', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 3, (1, 2)),),
                              ),
                 ))

    def test_put_instance_08(self):
        # commit inside record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(200):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
            self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=200,
                                    Game__bits=0,
                                    Game__list=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 200, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 201)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 201)])),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*24,
                                                     b'\x80',
                                                     b'\x00'*8166)),),
                              Black=(('B Smith', 0, 200, (1, 2)),),
                              ),
                 ))

    def test_put_instance_09(self):
        # commit after record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(2000):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=2000,
                                    Game__bits=2,
                                    Game__list=0,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=2,
                                    ),
                 records=dict(White=(('A Jones', 0, 2000, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*249,
                                                    b'\x80',
                                                    b'\x00'*7941)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*249,
                                                    b'\x80',
                                                    b'\x00'*7941)),
                                          ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*249,
                                                     b'\x80',
                                                     b'\x00'*7941)),
                                           ),
                              Black=(('B Smith', 0, 2000, (1, 2)),),
                              ),
                 ))

    def test_put_instance_10(self):
        # commit after record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(66000):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66000,
                                    Game__bits=2,
                                    Game__list=2,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=1,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 465, (1, 2, 3)),
                                     ),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*8191)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*8191)),
                                          ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in range(465)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(465)]),
                                          ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*8191)),
                                           b''.join((b'\xff'*58,
                                                     b'\x80',
                                                     b'\x00'*8133)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 465, (1, 2, 3)),
                                     ),
                              ___control=(('L', 2),
                                          ),
                              ),
                 ))


class DBbitapi_delete_instance(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('delete_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('delete_instance')

    def test_delete_instance_01(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # Do a put_instance test to establish starting point.
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=3,
                                    Game__bits=0,
                                    Game__list=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 3, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 4)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 4)])),
                              Game__exist=(b''.join((b'\x70', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 3, (1, 2)),),
                              ),
                 ))
        # Do the delete_instance test.
        # Delete record number 2
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 2))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=2,
                                    Game__bits=0,
                                    Game__list=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,#1,
                                    ),
                 records=dict(White=(('A Jones', 0, 2, (1, 2)),),
                              Game=dict(null={2},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in (1, 3)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in (1, 3)])),
                              Game__exist=(b''.join((b'\x50', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 2, (1, 2)),),
                              #___control=(('E', 0),
                              #            ),
                              ),
                 ))
        # Delete record number 3
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 3))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=1,
                                    Game__bits=0,
                                    Game__list=0,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=2,#3,
                                    ),
                 records=dict(White=(('A Jones', 0, 1, (1, 2)),),
                              Game=dict(null={2, 3},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__exist=(b''.join((b'\x40', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 1, (1, 2)),),
                              ___control=(#('E', 0),
                                          ('L', 1),
                                          ('L', 2),
                                          ),
                              ),
                 ))

    def test_delete_instance_02(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # Do a put_instance test to establish starting point.
        # Exactly the number of records so second delete from segment number 1
        # should convert segment bitmap to list of record numbers:
        # range(65536 + DB_CONVERSION_LIMIT + 1).
        self.api.open_context()
        for i in range(66048):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66048,
                                    Game__bits=4,
                                    Game__list=0,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=3,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 513, (3, 4)),
                                     ),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*8191)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*8191)),
                                          b''.join((b'\xff'*64,
                                                    b'\x80',
                                                    b'\x00'*8127)),
                                          b''.join((b'\xff'*64,
                                                    b'\x80',
                                                    b'\x00'*8127)),
                                          ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*8191)),
                                           b''.join((b'\xff'*64,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 513, (3, 4)),
                                     ),
                              ___control=(('L', 1),
                                          ('L', 2),
                                          ('L', 3),
                                          ),
                              ),
                 ))
        # Do the delete_instance test.
        # Delete record number 58
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 58))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66047,
                                    Game__bits=4,
                                    Game__list=0,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=4,
                                    ),
                 records=dict(White=(('A Jones', 0, 65534, (1, 2)),
                                     ('A Jones', 1, 513, (3, 4)),
                                     ),
                              Game=dict(null={58},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xdf',
                                                    b'\xff'*8184)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xdf',
                                                    b'\xff'*8184)),
                                          b''.join((b'\xff'*31,
                                                    b'\xff',
                                                    b'\xff'*32,
                                                    b'\x80',
                                                    b'\x00'*8127)),
                                          b''.join((b'\xff'*31,
                                                    b'\xff',
                                                    b'\xff'*32,
                                                    b'\x80',
                                                    b'\x00'*8127)),
                                          ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xdf',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xff',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65534, (1, 2)),
                                     ('B Smith', 1, 513, (3, 4)),
                                     ),
                              ___control=(('E', 0),
                                          ('L', 1),
                                          ('L', 2),
                                          ('L', 3),
                                          ),
                              ),
                 ))
        # Delete record number 65790
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65790))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66046,
                                    Game__bits=4,
                                    Game__list=0,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=4,#5,
                                    ),
                 records=dict(White=(('A Jones', 0, 65534, (1, 2)),
                                     ('A Jones', 1, 512, (3, 4)),
                                     ),
                              Game=dict(null={58, 65790},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xdf',
                                                    b'\xff'*8184)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xdf',
                                                    b'\xff'*8184)),
                                          b''.join((b'\xff'*31,
                                                    b'\xfd',
                                                    b'\xff'*32,
                                                    b'\x80',
                                                    b'\x00'*8127)),
                                          b''.join((b'\xff'*31,
                                                    b'\xfd',
                                                    b'\xff'*32,
                                                    b'\x80',
                                                    b'\x00'*8127)),
                                          ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xdf',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xfd',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65534, (1, 2)),
                                     ('B Smith', 1, 512, (3, 4)),
                                     ),
                              ___control=(('E', 0),
                                          #('E', 1),
                                          ('L', 1),
                                          ('L', 2),
                                          ('L', 3),
                                          ),
                              ),
                 ))
        # Delete record number 65787
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65787))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66045,
                                    Game__bits=2,
                                    Game__list=2,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=5,#6,
                                    ),
                 records=dict(White=(('A Jones', 0, 65534, (1, 2)),
                                     ('A Jones', 1, 511, (3, 4)),
                                     ),
                              Game=dict(null={58, 65790, 65787},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xdf',
                                                    b'\xff'*8184)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xdf',
                                                    b'\xff'*8184)),
                                          ),
                              Game__list=(
                                  b''.join((
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(31*8)]),
                                      b''.join([(i+31*8).to_bytes(2, 'big')
                                                for i in
                                                (0, 1, 2, 4, 5, 7)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(
                                                    32*8, 1+64*8)]),
                                      )),
                                  b''.join((
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(31*8)]),
                                      b''.join([(i+31*8).to_bytes(2, 'big')
                                                for i in
                                                (0, 1, 2, 4, 5, 7)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(
                                                    32*8, 1+64*8)]),
                                      )),
                                  ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xdf',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65534, (1, 2)),
                                     ('B Smith', 1, 511, (3, 4)),
                                     ),
                              ___control=(('B', 3),
                                          ('B', 4),
                                          ('E', 0),
                                          #('E', 1),
                                          ('L', 2),
                                          ('L', 3),
                                          ),
                              ),
                 ))


class DBbitapi_edit_instance(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('edit_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('edit_instance')

    def test_edit_instance_01(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        editvalue = {WHITE_FIELD_DEF:'C Brown',
                     BLACK_FIELD_DEF:'D Grey',
                     }
        # set starting point with 3 records.
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=3,
                                    Game__bits=0,
                                    Game__list=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used.
                 # White will get one and Black the other, at random.
                 records=dict(White=(('A Jones', 0, 3, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 4)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 4)])),
                              Game__exist=(b''.join((b'\x70', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 3, (1, 2)),),
                              ),
                 ))
        self.api.open_context()
        # Edit record number 2.
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 2))
        er = self.sr.clone()
        er.value.White = editvalue[WHITE_FIELD_DEF]
        er.value.Black = editvalue[BLACK_FIELD_DEF]
        self.sr.newrecord = er
        self.api.edit_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=3,
                                    Game__bits=0,
                                    Game__list=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used.
                 # White will get one and Black the other, at random.
                 # Brown and Grey are direct record numbers (one record only).
                 records=dict(White=(('A Jones', 0, 2, (1, 2)),
                                     ('C Brown', 0, 1, (2,)),
                                     ),
                              Game=dict(null={},
                                        rows={1:dict(
                                            White='C Brown', Black='D Grey') },
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in (1, 3)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in (1, 3)])),
                              Game__exist=(b''.join((b'\x70', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 2, (1, 2)),
                                     ('D Grey', 0, 1, (2,)),
                                     ),
                              ),
                 ))

    def test_edit_instance_02(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        editvalue = {WHITE_FIELD_DEF:'C Brown',
                     BLACK_FIELD_DEF:'D Grey',
                     }
        # set starting point with 9 records.
        # To get lists of record numbers changed by third edit.
        self.api.open_context()
        for i in range(9):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=9,
                                    Game__bits=0,
                                    Game__list=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used.
                 # White will get one and Black the other, at random.
                 records=dict(White=(('A Jones', 0, 9, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__list=(b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 10)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(1, 10)])),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xc0',
                                                     b'\x00'*8190)),),
                              Black=(('B Smith', 0, 9, (1, 2)),),
                              ),
                 ))
        self.api.open_context()
        # Edit record numbers 2, 5, 6.
        for i in (2, 5, 6):
            self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, i))
            er = self.sr.clone()
            er.value.White = editvalue[WHITE_FIELD_DEF]
            er.value.Black = editvalue[BLACK_FIELD_DEF]
            self.sr.newrecord = er
            self.api.edit_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=9,
                                    Game__bits=0,
                                    Game__list=4,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used,
                 # White will get one and Black the other, at random.
                 # Brown and Grey get segment numbers (3, 4).
                 records=dict(White=(('A Jones', 0, 6, (1, 2)),
                                     ('C Brown', 0, 3, (3, 4)),
                                     ),
                              Game=dict(null={},
                                        rows={k:dict(White='C Brown',
                                                     Black='D Grey')
                                              for k in (1, 4, 5)},
                                        default=dict(White='A Jones',
                                                     Black='B Smith'),
                                        ),
                              Game__list=(
                                  b''.join([i.to_bytes(2, 'big')
                                            for i in (1, 3, 4, 7, 8, 9)]),
                                  b''.join([i.to_bytes(2, 'big')
                                            for i in (1, 3, 4, 7, 8, 9)]),
                                  b''.join([i.to_bytes(2, 'big')
                                            for i in (2, 5, 6)]),
                                  b''.join([i.to_bytes(2, 'big')
                                            for i in (2, 5, 6)]),
                                  ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xc0',
                                                     b'\x00'*8190)),),
                              Black=(('B Smith', 0, 6, (1, 2)),
                                     ('D Grey', 0, 3, (3, 4)),
                                     ),
                              ),
                 ))


class DBbitapi_put_instance_with_holes(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('put_instance_holes')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('put_instance_holes')

    def test_holes_01(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # set starting point like end of test_delete_instance_05
        self.api.open_context()
        for i in range(66048):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 58))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65790))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65787))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66045,
                                    Game__bits=2,
                                    Game__list=2,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=5,
                                    ),
                 records=dict(White=(('A Jones', 0, 65534, (1, 2)),
                                     ('A Jones', 1, 511, (3, 4)),
                                     ),
                              Game=dict(null={58, 65790, 65787},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xdf',
                                                    b'\xff'*8184)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xdf',
                                                    b'\xff'*8184)),
                                          ),
                              Game__list=(
                                  b''.join((
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(31*8)]),
                                      b''.join([(i+31*8).to_bytes(2, 'big')
                                                for i in
                                                (0, 1, 2, 4, 5, 7)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(
                                                    32*8, 1+64*8)]),
                                      )),
                                  b''.join((
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(31*8)]),
                                      b''.join([(i+31*8).to_bytes(2, 'big')
                                                for i in
                                                (0, 1, 2, 4, 5, 7)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(
                                                    32*8, 1+64*8)]),
                                      )),
                                  ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xdf',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65534, (1, 2)),
                                     ('B Smith', 1, 511, (3, 4)),
                                     ),
                              ___control=(('B', 3),
                                          ('B', 4),
                                          ('E', 0),
                                          ('L', 2),
                                          ('L', 3),
                                          ),
                              ),
                 ))
        # Insert a record
        self.api.open_context()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66046,
                                    Game__bits=2,
                                    Game__list=2,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=5,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 511, (3, 4)),
                                     ),
                              Game=dict(null={65790, 65787},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xff',#df',
                                                    b'\xff'*8184)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xff',#df',
                                                    b'\xff'*8184)),
                                          ),
                              Game__list=(
                                  b''.join((
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(31*8)]),
                                      b''.join([(i+31*8).to_bytes(2, 'big')
                                                for i in
                                                (0, 1, 2, 4, 5, 7)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(
                                                    32*8, 1+64*8)]),
                                      )),
                                  b''.join((
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(31*8)]),
                                      b''.join([(i+31*8).to_bytes(2, 'big')
                                                for i in
                                                (0, 1, 2, 4, 5, 7)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(
                                                    32*8, 1+64*8)]),
                                      )),
                                  ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xff',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 511, (3, 4)),
                                     ),
                              ___control=(('B', 3),
                                          ('B', 4),
                                          ('E', 0),
                                          ('L', 2),
                                          ('L', 3),
                                          ),
                              ),
                 ))
        # Fill the high segment
        self.api.open_context()
        for i in range(1+64*8, 65536):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=131069,
                                    Game__bits=4,
                                    Game__list=0,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=5,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 65534, (3, 4)),
                                     ),
                              Game=dict(null={65790, 65787},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xff',
                                                    b'\xff'*8184)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xff',
                                                    b'\xff'*8184)),
                                          # Segment 1 back to bitmap form.
                                          b''.join((b'\xff'*31,
                                                    b'\xed',
                                                    b'\xff'*32,
                                                    b'\xff',
                                                    b'\xff'*8127)),
                                          b''.join((b'\xff'*31,
                                                    b'\xed',
                                                    b'\xff'*32,
                                                    b'\xff',
                                                    b'\xff'*8127)),
                                          ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xff',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\xff',
                                                     b'\xff'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 65534, (3, 4)),
                                     ),
                              ___control=(('B', 4),
                                          ('L', 1),
                                          ('L', 2),
                                          ('L', 3),
                                          ('L', 4),
                                          ),
                              ),
                 ))
        # Delete a record to set reuse for segment 1.
        self.api.open_context()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 66053))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 131072))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=131068,
                                    Game__bits=4,
                                    Game__list=0,
                                    Game__exist=3,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=6,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 65533, (3, 4)),
                                     ),
                              Game=dict(null={65790, 65787, 66053, 131072},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xff',
                                                    b'\xff'*8184)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xff',
                                                    b'\xff'*8184)),
                                          b''.join((b'\xff'*31,
                                                    b'\xed',
                                                    b'\xff'*32,
                                                    b'\xfb',
                                                    b'\xff'*8127)),
                                          b''.join((b'\xff'*31,
                                                    b'\xed',
                                                    b'\xff'*32,
                                                    b'\xfb',
                                                    b'\xff'*8127)),
                                          ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xff',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\xfb',
                                                     b'\xff'*8127)),
                                           b''.join((b'\x00',
                                                     b'\x00'*8191)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 65533, (3, 4)),
                                     ),
                              ___control=(('B', 4),
                                          ('E', 1),
                                          ('L', 1),
                                          ('L', 2),
                                          ('L', 3),
                                          ('L', 4),
                                          ),
                              ),
                 ))
        # Insert three records.
        self.api.open_context()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=131071,
                                    Game__bits=4,
                                    Game__list=0,
                                    Game__exist=3,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=6,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 65536, (3, 4)),
                                     ),
                              Game=dict(null={131072},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__bits=(b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xff',
                                                    b'\xff'*8184)),
                                          b''.join((b'\x7f',
                                                    b'\xff'*6,
                                                    b'\xff',
                                                    b'\xff'*8184)),
                                          b''.join((b'\xff'*31,
                                                    b'\xff',
                                                    b'\xff'*32,
                                                    b'\xff',
                                                    b'\xff'*8127)),
                                          b''.join((b'\xff'*31,
                                                    b'\xff',
                                                    b'\xff'*32,
                                                    b'\xff',
                                                    b'\xff'*8127)),
                                          ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xff',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xff',
                                                     b'\xff'*32,
                                                     b'\xff',
                                                     b'\xff'*8127)),
                                           b''.join((b'\x00',
                                                     b'\x00'*8191)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 65536, (3, 4)),
                                     ),
                              ___control=(('B', 4),
                                          ('E', 1),
                                          ('L', 1),
                                          ('L', 2),
                                          ('L', 3),
                                          ('L', 4),
                                          ),
                              ),
                 ))


def record_counts(testcase, db_directory, answers):
    list_seg_length = constants.LENGTH_SEGMENT_LIST_REFERENCE
    bits_seg_length = constants.LENGTH_SEGMENT_BITARRAY_REFERENCE
    files = dict(Round='games_Round',
                 White='games_White',
                 Event='games_Event',
                 Site='games_Site',
                 Black='games_Black',
                 Date='games_Date',
                 Result='games_Result',
                 )
    literal_eval = ast.literal_eval
    testcase.assertEqual(
        bool(set(answers['records']) - set(answers['record_counts'])),
        False)
    dbenv = DBEnv()
    dbenv.open(db_directory)
    for t in ('Round',
              'White',
              'Game',
              'Game__bits',
              'Game__list',
              'Game__exist',
              'Event',
              'Site',
              'Black',
              'Date',
              'Result',
              '___control',
              ):
        db = DB(dbenv)
        db.open(files.get(t, t), dbname=t, flags=DB_RDONLY)
        stat = db.stat()
        testcase.assertEqual(stat['ndata'], answers['record_counts'][t])
        db.close()
    # Collect the values like in sqlite3 tests, but the testable constraint
    # is not yet clear.
    segments = set()
    rlast = None
    for k, v in answers['records'].items():
        db = DB(dbenv)
        db.open(files.get(k, k), dbname=k, flags=DB_RDONLY)
        cursor = db.cursor()
        # Berkeley DB records are 1-base but tuple indexes are 0-base.
        row = 0
        r = cursor.first()
        while r:
            row += 1
            rk, rv = r
            if k == 'Game':
                # Deleted records are represented by an empty row in sqlite3
                # but really are deleted in bsddb3 databases. This allows the
                # segment slot for a deleted record to be reused.
                testcase.assertEqual(rk, row + len([i for i in v['null']
                                                    if rk > i]))
                testcase.assertEqual(
                    literal_eval(rv.decode()),
                    v['rows'].get(row-1, v['default']))
            elif k in ('White', 'Black'):
                testcase.assertEqual(rk.decode(), v[row-1][0])
                if len(r[1]) == list_seg_length:
                    # List of record numbers.
                    srn = int.from_bytes(rv[6:], byteorder='big')
                    seg = int.from_bytes(rv[:4], byteorder='big')
                    rct = int.from_bytes(rv[4:6], byteorder='big')
                    segments.add(srn)
                elif len(r[1]) == bits_seg_length:
                    # Bitmap of record numbers.
                    srn = int.from_bytes(rv[7:], byteorder='big')
                    seg = int.from_bytes(rv[:4], byteorder='big')
                    rct = int.from_bytes(rv[4:7], byteorder='big')
                    segments.add(srn)
                else:
                    # Direct record number.
                    srn = int.from_bytes(rv[4:], byteorder='big')
                    seg = int.from_bytes(rv[:4], byteorder='big')
                    rct = 1
                testcase.assertEqual((rk.decode(), seg, rct), v[row-1][:-1])
                # Cannot be sure which value got used, but 'White' and
                # 'Black' must use a different one of those given.
                #testcase.assertEqual(srn in v[row-1][-1], True)
            elif k == 'Game__exist':
                testcase.assertEqual(rk, row)
                testcase.assertEqual(rv, v[row-1])
            elif k == 'Game__bits':
                testcase.assertEqual(rv, v[row-1])
            elif k == 'Game__list':
                testcase.assertEqual(rv, v[row-1])
            elif k == '___control':
                # Queues of released segments not yet re-used.
                testcase.assertEqual(rk.decode(), v[row-1][0])
                testcase.assertEqual(
                    int.from_bytes(rv, byteorder='big'),
                    v[row-1][-1])
            r = cursor.next()
    dbenv.close()


if __name__ == '__main__':
    runner = unittest.TextTestRunner
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    
    runner().run(loader(SampleRecordAssumptions))
    
    runner().run(loader(_DatabaseEncoders))

    runner().run(loader(_DBapi_init))
    runner().run(loader(_DBapi_backout))
    runner().run(loader(_DBapi_close_context))
    runner().run(loader(_DBapi_close_contexts))
    runner().run(loader(_DBapi_close_database))
    runner().run(loader(_DBapi_commit))
    runner().run(loader(_DBapi_db_compatibility_hack))
    runner().run(loader(_DBapi_do_deferred_updates))
    runner().run(loader(_DBapi_exists))
    runner().run(loader(_DBapi_files_exist))
    runner().run(loader(_DBapi_database_cursor))
    runner().run(loader(_DBapi_repair_cursor))
    runner().run(loader(_DBapi_get_database_folder))
    runner().run(loader(_DBapi_get_database))
    runner().run(loader(_DBapi_get_database_instance))
    runner().run(loader(_DBapi_get_first_primary_key_for_index_key))
    runner().run(loader(_DBapi_get_primary_record))
    runner().run(loader(_DBapi_is_primary))
    runner().run(loader(_DBapi_is_primary_recno))
    runner().run(loader(_DBapi_is_recno))
    runner().run(loader(_DBapi_open_contexts))
    runner().run(loader(_DBapi_allocate_and_open_contexts))
    runner().run(loader(_DBapi_get_packed_key))
    runner().run(loader(_DBapi_decode_as_primary_key))
    runner().run(loader(_DBapi_encode_primary_key))
    runner().run(loader(_DBapi_set_defer_update))
    runner().run(loader(_DBapi_unset_defer_update))
    runner().run(loader(_DBapi_use_deferred_update_process))
    runner().run(loader(_DBapi_initial_database_size))
    runner().run(loader(_DBapi_increase_database_size))
    runner().run(loader(_DBapi_do_database_task))
    runner().run(loader(_DBapi_make_recordset_key))
    runner().run(loader(_DBapi_make_recordset_key_startswith))
    runner().run(loader(_DBapi_make_recordset_key_range))
    runner().run(loader(_DBapi_make_recordset_all))
    runner().run(loader(_DBapi_recordset_for_segment))
    runner().run(loader(_DBapi_file_records_under))
    runner().run(loader(_DBapi_is_engine_uses_bytes))
    runner().run(loader(_DBapi_is_engine_uses_str))
    runner().run(loader(_DBapi_start_transaction))
    runner().run(loader(_DBapi_cede_contexts_to_process))

    runner().run(loader(DBbitapi_method_classes))
    runner().run(loader(DBbitapi_init))
    runner().run(loader(DBbitapi_encode_record_key))
    runner().run(loader(DBbitapi_open_context))
    runner().run(loader(DBbitapi_close_context))
    runner().run(loader(DBbitapi_put_instance))
    runner().run(loader(DBbitapi_delete_instance))
    runner().run(loader(DBbitapi_edit_instance))

    runner().run(loader(DBbitapi_put_instance_with_holes))
