# sqlite3api_test.py
# Copyright 2014 Roger Marsh
# License: See LICENSE.TXT (BSD licence)

"""sqlite3api tests"""

import unittest
import os
import shutil
import sqlite3
import subprocess
import sys
import ast

from ..api.test._filespec import (
    samplefilespec,
    SampleValue,
    SampleRecord,
    GAMES_FILE_DEF,
    GAME_FIELD_DEF,
    EVENT_FIELD_DEF,
    SITE_FIELD_DEF,
    DATE_FIELD_DEF,
    ROUND_FIELD_DEF,
    WHITE_FIELD_DEF,
    BLACK_FIELD_DEF,
    RESULT_FIELD_DEF,
    SEVEN_TAG_ROSTER,
    )
from .. import sqlite3api
from ..api import filespec, record, recordset, database
from ..api.constants import (
    DB_UPPER_CONVERSION_LIMIT,
    DB_LOWER_CONVERSION_LIMIT,
    )


def _api(folder):
    return sqlite3api._Sqlite3api(
        sqlite3api.Sqlite3bitPrimary,
        sqlite3api.Sqlite3bitSecondary,
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'sqlite3api_tests', folder)))


def api(folder):
    return sqlite3api.Sqlite3api(
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'sqlite3api_tests', folder)))


def bitapi(folder):
    return sqlite3api.Sqlite3bitapi(
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'sqlite3api_tests', folder)))


def _delete_folder(folder):
    shutil.rmtree(
        os.path.expanduser(os.path.join('~', 'sqlite3api_tests', folder)),
        ignore_errors=True)


# Defined at module level for pickling.
class _Value(record.Value):
    def pack(self):
        v = super().pack()
        v[1]['Site'] = 'gash' # defined in samplefilespec()
        v[1]['hhhhhh'] = 'hhhh' # not defined in samplefilespec(), ignored
        return v


# Defined at module level for pickling.
class _ValueEdited(record.Value):
    def pack(self):
        v = super().pack()
        v[1]['Site'] = 'newgash' # defined in samplefilespec()
        v[1]['hhhhhh'] = 'hhhh' # not defined in samplefilespec(), ignored
        return v


class _DatabaseEncoders(unittest.TestCase):

    def setUp(self):
        self.sde = sqlite3api._DatabaseEncoders()

    def tearDown(self):
        pass

    def test____assumptions(self):
        self.assertEqual(repr(57), '57')

    def test_encode_record_number_01(self):
        self.assertEqual(self.sde.encode_record_number(57), '57')

    def test_decode_record_number_01(self):
        self.assertRaisesRegex(
            ValueError,
            "malformed node or string: 57",
            self.sde.decode_record_number,
            57)
        self.assertEqual(self.sde.decode_record_number('57'), 57)

    def test_encode_record_selector(self):
        self.assertEqual(self.sde.encode_record_selector('57'), '57')


class _Sqlite3api_init(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        p = os.path.abspath('')
        self.assertEqual(
            sqlite3api._Sqlite3api(None, None, {}, '').__dict__,
            dict(_sqtables={},
                 _sqfile=os.path.join(p, os.path.basename(p)),
                 _sqconn=None,
                 _associate={}))
        self.assertIs(
            sqlite3api._Sqlite3api(None, None, {}, '').engine_uses_bytes_or_str,
            sqlite3api._Sqlite3api.engine_uses_bytes_or_str)
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 4 required positional arguments: ",
                "'primary_class', 'secondary_class', 'sqlite3tables', and ",
                "'sqlite3databasefolder'",
                )),
            sqlite3api._Sqlite3api,
            )
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Database file name None is not valid",
            sqlite3api._Sqlite3api,
            *(None, None, None, None, None))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Table definitions must be a dictionary",
            sqlite3api._Sqlite3api,
            *(None, None, None, '', None))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Table definition for 't' must be a dictionary",
            sqlite3api._Sqlite3api,
            *(None, None, dict(t=None), '', None))
        self.assertRaisesRegex(
            KeyError,
            "'primary'",
            sqlite3api._Sqlite3api,
            *(None, None, dict(t={}), '', None))
        self.assertRaisesRegex(
            KeyError,
            "'fields'",
            sqlite3api._Sqlite3api,
            *(None,
              None,
              dict(t=dict(primary='key')),
              ''))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Primary column name key for t does not have a column description",
            sqlite3api._Sqlite3api,
            *(None,
              None,
              dict(t=dict(primary='key', fields=('lock',))),
              ''))
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not callable",
            sqlite3api._Sqlite3api,
            *(None,
              None,
              dict(t=dict(primary='key', fields=('lock', 'key'))),
              ''))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "".join((
                "Attributes for column 'key' in table 'key' must be a 'dict' ",
                "or 'None'")),
            sqlite3api._Sqlite3api,
            *(sqlite3api.Sqlite3bitPrimary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key=True),
                      filedesc=dict(fileorg=None))),
              ''))
        self.assertIsInstance(
            sqlite3api._Sqlite3api(
                sqlite3api.Sqlite3bitPrimary,
                None,
                dict(
                    t=dict(
                        primary='key',
                        fields=dict(lock={}, key={}),
                        filedesc=dict(fileorg=None))),
                ''),
            sqlite3api._Sqlite3api)
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'items'",
            sqlite3api._Sqlite3api,
            *(sqlite3api.Sqlite3bitPrimary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary=None)),
              ''))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Secondary table name 1 for 't' must be a string",
            sqlite3api._Sqlite3api,
            *(sqlite3api.Sqlite3bitPrimary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary={1:2})),
              ''))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Secondary table name key for t already used",
            sqlite3api._Sqlite3api,
            *(sqlite3api.Sqlite3bitPrimary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary=dict(key='key'),
                      )),
              ''))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Secondary table name key for t already used",
            sqlite3api._Sqlite3api,
            *(sqlite3api.Sqlite3bitPrimary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary=dict(lock='key'),
                      )),
              ''))

        # Now must have FileSpec which brings in some DPT related stuff.
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Secondary table name Key for t does not have a column description",
            sqlite3api._Sqlite3api,
            *(sqlite3api.Sqlite3bitPrimary,
              None,
              filespec.FileSpec(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None, brecppg=100),
                      secondary=dict(key=None),
                      btod_factor=5,
                      )),
              ''))
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not callable",
            sqlite3api._Sqlite3api,
            *(sqlite3api.Sqlite3bitPrimary,
              None,
              filespec.FileSpec(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}, Key={}),
                      filedesc=dict(fileorg=None, brecppg=100),
                      secondary=dict(key=None),
                      btod_factor=5,
                      )),
              ''))

        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not callable",
            sqlite3api._Sqlite3api,
            *(sqlite3api.Sqlite3bitPrimary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary=dict(key='lock'),
                      )),
              ''))

    def test___init___02(self):
        fs = dict(
            t=dict(
                primary='key',
                fields=dict(lock={}, key={}),
                filedesc=dict(fileorg=None),
                secondary=dict(key='lock'),
                ))
        api = sqlite3api._Sqlite3api(
            sqlite3api.Sqlite3bitPrimary,
            sqlite3api.Sqlite3bitSecondary,
            dict(
                t=dict(
                    primary='key',
                    fields=dict(lock={}, key={}),
                    filedesc=dict(fileorg=None),
                    secondary=dict(key='lock'),
                    )),
            '')
        p = os.path.abspath('')
        self.assertEqual(api._sqfile, os.path.join(p, os.path.basename(p)))
        self.assertEqual(api._sqconn, None)
        self.assertEqual(api._associate, dict(t=dict(key='lock', t='key')))
        self.assertEqual(len(api._sqtables), 2)
        self.assertIsInstance(
            api._sqtables['key'], sqlite3api.Sqlite3bitPrimary)
        self.assertIsInstance(
            api._sqtables['lock'], sqlite3api.Sqlite3bitSecondary)
        self.assertEqual(len(api.__dict__), 4)


class _Sqlite3api_open_context(unittest.TestCase):

    def setUp(self):
        self.api = _api('open_context')

    def tearDown(self):
        _delete_folder('open_context')

    def test_open_context_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(os.path.exists(self.api._sqfile), False)
        self.assertEqual(self.api.open_context(), True)
        conn = self.api._sqconn
        self.assertIsInstance(conn, sqlite3.Connection)
        self.assertEqual(os.path.isfile(self.api._sqfile), True)
        self.assertEqual(self.api.open_context(), True)
        self.assertIs(self.api._sqconn, conn)
        self.assertEqual(os.path.isfile(self.api._sqfile), True)


class _Sqlite3api_close_context(unittest.TestCase):

    def setUp(self):
        self.api = _api('close_context')

    def tearDown(self):
        _delete_folder('close_context')

    def test_close_context_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.close_context(), None)
        self.assertEqual(self.api._sqconn, None)

    def test_close_context_02(self):
        self.api.open_context()
        conn = self.api._sqconn
        self.assertIsInstance(conn, sqlite3.Connection)
        self.assertEqual(self.api.close_context(), None)
        self.assertIs(self.api._sqconn, conn)
        self.assertEqual(self.api.close_context(), None)
        self.assertIs(self.api._sqconn, conn)


class _Sqlite3api_close_database(unittest.TestCase):

    def setUp(self):
        self.api = _api('close_database')

    def tearDown(self):
        _delete_folder('close_database')

    def test_close_database_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.close_database(), None)
        self.assertEqual(self.api._sqconn, None)

    def test_close_database_02(self):
        self.api.open_context()
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)
        self.assertEqual(self.api.close_database(), None)
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.close_database(), None)
        self.assertEqual(self.api._sqconn, None)


class _Sqlite3api_backout(unittest.TestCase):
    # Nothing to backout, just check the calls work.

    def setUp(self):
        self.api = _api('backout')

    def tearDown(self):
        _delete_folder('backout')

    def test_backout_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.backout(), None)
        self.api.open_context()
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)
        self.assertEqual(self.api.backout(), None)
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)


class _Sqlite3api_close_contexts(unittest.TestCase):

    def setUp(self):
        self.api = _api('close_contexts')

    def tearDown(self):
        _delete_folder('close_contexts')

    def test_close_contexts_01(self):
        self.assertEqual(self.api.close_contexts('DPT compatibility'), None)


class _Sqlite3api_commit(unittest.TestCase):
    # Nothing to commit, just check the calls work.

    def setUp(self):
        self.api = _api('commit')

    def tearDown(self):
        _delete_folder('commit')

    def test_commit_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.commit(), None)
        self.api.open_context()
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)
        self.assertEqual(self.api.commit(), None)
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)


class _Sqlite3api_db_compatibility_hack(unittest.TestCase):

    def setUp(self):
        self.api = _api('db_compatibility_hack')

    def tearDown(self):
        _delete_folder('db_compatibility_hack')

    def test_db_compatibility_hack_01(self):
        record = ('key', 'value')
        srkey = self.api.encode_record_number(3456)
        self.assertEqual(
            self.api.db_compatibility_hack(record, srkey), record)
        record = ('key', None)
        value = self.api.decode_record_number(srkey)
        self.assertEqual(
            self.api.db_compatibility_hack(record, srkey),
            ('key', value))


class _Sqlite3api_exists(unittest.TestCase):

    def setUp(self):
        self.api = _api('exists')

    def tearDown(self):
        _delete_folder('exists')

    def test_exists_01(self):
        self.assertEqual(self.api.exists(None, None), False)
        self.assertEqual(self.api.exists('games', None), False)
        self.assertEqual(self.api.exists('games', 'Site'), True)


class _Sqlite3api_files_exist(unittest.TestCase):

    def setUp(self):
        self.api = _api('files_exist')

    def tearDown(self):
        _delete_folder('files_exist')

    def test_exists_01(self):
        self.assertEqual(self.api.files_exist(), False)
        self.api.open_context()
        self.assertEqual(self.api.files_exist(), True)
        self.api.close_database()
        self.assertEqual(self.api.files_exist(), True)


class _Sqlite3api_make_cursor(unittest.TestCase):

    def setUp(self):
        self.api = _api('database_cursor')

    def tearDown(self):
        _delete_folder('database_cursor')

    def test_make_cursor_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.database_cursor,
            *(None, None))
        self.assertIsInstance(
            self.api.database_cursor('games', 'Site'),
            sqlite3api.CursorSqlite3bitSecondary)
        self.assertIsInstance(
            self.api.database_cursor('games', 'games'),
            sqlite3api.CursorSqlite3bitPrimary)


class _Sqlite3api_get_database_folder(unittest.TestCase):

    def setUp(self):
        self.api = _api('get_database_folder')

    def tearDown(self):
        _delete_folder('get_database_folder')

    def test_get_database_folder_01(self):
        self.assertEqual(
            self.api.get_database_folder(),
            os.path.expanduser(
                os.path.join('~', 'sqlite3api_tests', 'get_database_folder')))


class _Sqlite3api_get_database(unittest.TestCase):

    def setUp(self):
        self.api = _api('get_database')

    def tearDown(self):
        _delete_folder('get_database')

    def test_get_database_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_database,
            *(None, None))
        self.assertEqual(self.api.get_database('games', 'Site'), None)
        self.assertEqual(self.api.get_database('games', 'games'), None)
        self.api.open_context()
        self.assertIsInstance(
            self.api.get_database('games', 'Site'),
            sqlite3.Connection)
        self.assertIsInstance(
            self.api.get_database('games', 'games'),
            sqlite3.Connection)


class _Sqlite3api_get_database_instance(unittest.TestCase):

    def setUp(self):
        self.api = _api('get_database_instance')

    def tearDown(self):
        _delete_folder('get_database_instance')

    def test_get_database_instance_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_database_instance,
            *(None, None))
        self.assertIsInstance(
            self.api.get_database_instance('games', 'Site'),
            sqlite3api.Sqlite3bitSecondary)
        self.assertIsInstance(
            self.api.get_database_instance('games', 'games'),
            sqlite3api.Sqlite3bitPrimary)


class _Sqlite3api_get_first_primary_key_for_index_key(unittest.TestCase):
    # Test results are suspicious but may be correct.

    def setUp(self):
        self.api = _api('get_first_primary_key_for_index_key')

    def tearDown(self):
        _delete_folder('get_first_primary_key_for_index_key')

    def test_get_first_primary_key_for_index_key_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_first_primary_key_for_index_key,
            *(None, None, None))
        self.assertRaisesRegex(
            AttributeError,
            ''.join((
                "'NoneType' object has no attribute ",
                "'execute'")),
            self.api.get_first_primary_key_for_index_key,
            *('games', 'Site', 'k'))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "get_first_primary_key_for_index_key for primary table",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'games', 'k'))
        self.api.open_context()
        #self.assertEqual(
        #    self.api.get_first_primary_key_for_index_key('games', 'games', 'k'),
        #    None)
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "get_first_primary_key_for_index_key for primary table",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'games', 'k'))


class _Sqlite3api_get_primary_record(unittest.TestCase):

    def setUp(self):
        self.api = _api('get_primary_record')

    def tearDown(self):
        _delete_folder('get_primary_record')

    def test_get_primary_record_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_primary_record,
            *(None, None))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.get_primary_record,
            *('games', 'k'))
        self.api.open_context()
        self.assertEqual(self.api.get_primary_record('games', 'k'), None)


class _Sqlite3api_is_primary(unittest.TestCase):

    def setUp(self):
        self.api = _api('is_primary')

    def tearDown(self):
        _delete_folder('is_primary')

    def test_is_primary_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.is_primary,
            *(None, None))
        self.assertEqual(self.api.is_primary('games', 'games'), True)
        self.assertEqual(self.api.is_primary('games', 'Site'), False)


class _Sqlite3api_is_primary_recno(unittest.TestCase):

    def setUp(self):
        self.api = _api('is_primary_recno')

    def tearDown(self):
        _delete_folder('is_primary_recno')

    def test_is_primary_recno_01(self):
        self.assertEqual(
            self.api.is_primary_recno('Berkeley DB compatibility'), True)


class _Sqlite3api_is_recno(unittest.TestCase):

    def setUp(self):
        self.api = _api('is_recno')

    def tearDown(self):
        _delete_folder('is_recno')

    def test_is_recno_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.is_recno,
            *(None, None))
        self.assertEqual(self.api.is_recno('games', 'games'), True)
        self.assertEqual(self.api.is_recno('games', 'Site'), False)


class _Sqlite3api_open_contexts(unittest.TestCase):

    def setUp(self):
        self.api = _api('open_contexts')

    def tearDown(self):
        _delete_folder('open_contexts')

    def test_open_contexts_01(self):
        self.assertEqual(self.api.open_contexts('DPT compatibility'), None)


class _Sqlite3api_get_packed_key(unittest.TestCase):

    def setUp(self):
        self.api = _api('get_packed_key')

    def tearDown(self):
        _delete_folder('get_packed_key')

    def test_get_packed_key_01(self):
        instance = record.Record(record.Key, record.Value)
        self.assertEqual(
            self.api.get_packed_key('Berkeley DB compatibility', instance),
            instance.key.pack())


class _Sqlite3api_decode_as_primary_key(unittest.TestCase):

    def setUp(self):
        self.api = _api('decode_as_primary_key')

    def tearDown(self):
        _delete_folder('decode_as_primary_key')

    def test_decode_as_primary_key_01(self):
        self.assertEqual(self.api.decode_as_primary_key(
            'Ignored in sqlite3', 10), 10)
        srkey = self.api.encode_record_number(3456)
        self.assertEqual(self.api.decode_as_primary_key(
            'Ignored in sqlite3', srkey), 3456)


class _Sqlite3api_encode_primary_key(unittest.TestCase):

    def setUp(self):
        self.api = _api('encode_primary_key')

    def tearDown(self):
        _delete_folder('encode_primary_key')

    def test_encode_primary_key_01(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.load(23)
        self.assertEqual(
            self.api.encode_primary_key('games', instance),
            '23')


class _Sqlite3api_use_deferred_update_process(unittest.TestCase):

    def setUp(self):
        self.api = _api('use_deferred_update_process')

    def tearDown(self):
        _delete_folder('use_deferred_update_process')

    def test_use_deferred_update_process_01(self):
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "use_deferred_update_process not implemented",
            self.api.use_deferred_update_process,
            **{'zzz': 2, 'd': None})


class _Sqlite3api_initial_database_size(unittest.TestCase):
    # initial_database_size does nothing, it is for DPT compatibility.

    def setUp(self):
        self.api = _api('initial_database_size')

    def tearDown(self):
        _delete_folder('initial_database_size')

    def test_initial_database_size_01(self):
        self.assertEqual(self.api.initial_database_size(), True)


class _Sqlite3api_increase_database_size(unittest.TestCase):

    def setUp(self):
        self.api = _api('increase_database_size')

    def tearDown(self):
        _delete_folder('increase_database_size')

    def test_increase_database_size_01(self):
        self.assertEqual(
            self.api.increase_database_size(**{'l':'DPT compatibility'}), None)


class _Sqlite3api_do_database_task(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        _delete_folder('do_database_task')

    def test_do_database_task_01(self):
        def f(a, b, **c):
            pass
        self.api = _api('do_database_task')
        self.assertRaisesRegex(
            TypeError,
            ''.join((
                "__init__\(\) missing 3 required positional arguments: ",
                "'secondary_class', 'sqlite3tables', ",
                "and 'sqlite3databasefolder'",
                )),
            self.api.do_database_task,
            *(f,))

    def test_do_database_task_02(self):
        def f(a, b, **c):
            pass
        class C(sqlite3api._Sqlite3api):
            def __init__(self, *a, **k):
                super().__init__(
                    sqlite3api.Sqlite3Primary,
                    sqlite3api.Sqlite3Secondary,
                    samplefilespec(),
                    *a,
                    **k)
        self.api = C(
            os.path.expanduser(
                os.path.join('~', 'sqlite3api_tests', 'do_database_task')))
        self.assertEqual(self.api.do_database_task(f), None)


class Sqlite3bitapi_do_database_task(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        _delete_folder('do_database_task')

    def test_do_database_task_01(self):
        def f(a, b, **c):
            pass
        self.api = bitapi('do_database_task')
        self.assertRaisesRegex(
            TypeError,
            ''.join((
                "__init__\(\) missing 1 required positional argument: ",
                "'sqlite3databasefolder'",
                )),
            self.api.do_database_task,
            *(f,))

    def test_do_database_task_02(self):
        def f(a, b, **c):
            pass
        class C(sqlite3api._Sqlite3api):
            def __init__(self, *a, **k):
                super().__init__(
                    sqlite3api.Sqlite3Primary,
                    sqlite3api.Sqlite3Secondary,
                    samplefilespec(),
                    *a,
                    **k)
        self.api = C(
            os.path.expanduser(
                os.path.join('~', 'sqlite3api_tests', 'do_database_task')))
        self.assertEqual(self.api.do_database_task(f), None)


class _Sqlite3api_make_recordset_key(unittest.TestCase):

    def setUp(self):
        self.api = _api('make_recordset_key')

    def tearDown(self):
        _delete_folder('make_recordset_key')

    def test_make_recordset_key_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_key,
            *(None, None))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.make_recordset_key,
            *('games', 'games'))
        self.api.open_context()
        self.assertRaisesRegex(
            TypeError,
            "object of type 'NoneType' has no len\(\)",
            self.api.make_recordset_key,
            *('games', 'games'))
        self.assertIsInstance(
            self.api.make_recordset_key('games', 'Site'),
            recordset.Recordset)


class _Sqlite3api_make_recordset_key_startswith(unittest.TestCase):

    def setUp(self):
        self.api = _api('make_recordset_key_startswith')

    def tearDown(self):
        _delete_folder('make_recordset_key_startswith')

    def test_make_recordset_key_startswith_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_key_startswith,
            *(None, None))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            ''.join((
                "populate_recordset_key_startswith not available on ",
                "primary database")),
            self.api.make_recordset_key_startswith,
            *('games', 'games'))
        if sys.version_info[:2] < (3,4):
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected bytes, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        else:
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected a bytes-like object, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        self.api.open_context()
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            ''.join((
                "populate_recordset_key_startswith not available on ",
                "primary database")),
            self.api.make_recordset_key_startswith,
            *('games', 'games'))
        if sys.version_info[:2] < (3,4):
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected bytes, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        else:
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected a bytes-like object, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))


class Sqlite3bitapi_make_recordset_key_startswith(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('make_recordset_key_startswith')

    def tearDown(self):
        _delete_folder('make_recordset_key_startswith')

    def test_make_recordset_key_startswith_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_key_startswith,
            *(None, None))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            ''.join((
                "populate_recordset_key_startswith not available on ",
                "primary database")),
            self.api.make_recordset_key_startswith,
            *('games', 'games'))
        if sys.version_info[:2] < (3,4):
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected bytes, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        else:
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected a bytes-like object, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        self.api.open_context()
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            ''.join((
                "populate_recordset_key_startswith not available on ",
                "primary database")),
            self.api.make_recordset_key_startswith,
            *('games', 'games'))
        if sys.version_info[:2] < (3,4):
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected bytes, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        else:
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected a bytes-like object, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))


class _Sqlite3api_make_recordset_key_range(unittest.TestCase):

    def setUp(self):
        self.api = _api('make_recordset_key_range')

    def tearDown(self):
        _delete_folder('make_recordset_key_range')

    def test_make_recordset_key_range_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_key_range,
            *(None, None))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.make_recordset_key_range,
            *('games', 'games'))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.make_recordset_key_range,
            *('games', 'Site'))
        self.api.open_context()
        self.assertIsInstance(
            self.api.make_recordset_key_range('games', 'games'),
            recordset.Recordset)
        self.assertIsInstance(
            self.api.make_recordset_key_range('games', 'Site'),
            recordset.Recordset)


class _Sqlite3api_make_recordset_all(unittest.TestCase):

    def setUp(self):
        self.api = _api('make_recordset_all')

    def tearDown(self):
        _delete_folder('make_recordset_all')

    def test_make_recordset_all_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_all,
            *(None, None))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.make_recordset_all,
            *('games', 'games'))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.make_recordset_all,
            *('games', 'Site'))
        self.api.open_context()
        self.assertIsInstance(
            self.api.make_recordset_all('games', 'games'),
            recordset.Recordset)
        self.assertIsInstance(
            self.api.make_recordset_all('games', 'Site'),
            recordset.Recordset)


class _Sqlite3api_recordset_for_segment(unittest.TestCase):

    def setUp(self):
        self.api = _api('recordset_for_segment')

    def tearDown(self):
        _delete_folder('recordset_for_segment')

    def test_recordset_for_segment_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'dbset'",
            self.api.recordset_for_segment,
            *(None, None, None))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.recordset_for_segment,
            *(rs, None, None))
        self.assertIsInstance(
            self.api.recordset_for_segment(
                rs, 'games', (None, 5, 1, 20000)),
            recordset.Recordset)
        self.assertIsInstance(
            self.api.recordset_for_segment(
                rs, 'Site', (None, 6, 1, 0)),
            recordset.Recordset)
        self.api.open_context()
        self.assertIsInstance(
            self.api.recordset_for_segment(
                rs, 'games', (None, 1, 1, 65535)),
            recordset.Recordset)
        self.assertIsInstance(
            self.api.recordset_for_segment(
                rs, 'Site', (None, 2, 1, 20001)),
            recordset.Recordset)


class Sqlite3bitapi_recordset_for_segment(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('recordset_for_segment')

    def tearDown(self):
        _delete_folder('recordset_for_segment')

    def test_recordset_for_segment_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'dbset'",
            self.api.recordset_for_segment,
            *(None, None, None))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.recordset_for_segment,
            *(rs, None, None))
        self.assertIsInstance(
            self.api.recordset_for_segment(
                rs, 'games', (None, 5, 1, 20000)),
            recordset.Recordset)
        self.assertIsInstance(
            self.api.recordset_for_segment(
                rs, 'Site', (None, 6, 1, 0)),
            recordset.Recordset)
        self.api.open_context()
        self.assertIsInstance(
            self.api.recordset_for_segment(
                rs, 'games', (None, 1, 1, 65535)),
            recordset.Recordset)
        self.assertIsInstance(
            self.api.recordset_for_segment(
                rs, 'Site', (None, 2, 1, 20001)),
            recordset.Recordset)

    def test_recordset_for_segment_02(self):
        # As close as it gets without a segment index to read when number of
        # records in segment > 1
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.recordset_for_segment,
            *(rs, 'games', (None, 5, 2, 100)))

    def test_recordset_for_segment_03(self):
        # As close as it gets without a segment index to read when number of
        # records in segment > 1
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.recordset_for_segment,
            *(rs, 'games', (None, 5, DB_UPPER_CONVERSION_LIMIT + 10, 100)))


class _Sqlite3api_file_records_under(unittest.TestCase):

    def setUp(self):
        self.api = _api('file_records_under')

    def tearDown(self):
        _delete_folder('file_records_under')

    def test_file_records_under_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'dbidentity'",
            self.api.file_records_under,
            *(None, None, None, None))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.file_records_under,
            *(None, None, rs, None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "file_records_under not implemented for Sqlite3bitPrimary",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.file_records_under,
            *('games', 'Site', rs, b'dd'))
        self.api.open_context()
        self.assertRaisesRegex(
            database.DatabaseError,
            "Record set was not created from this database instance",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            database.DatabaseError,
            "file_records_under not implemented for Sqlite3bitPrimary",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs._dbset = 'sssss'
        self.assertRaisesRegex(
            database.DatabaseError,
            "Record set was not created from dbset database",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertEqual(
            self.api.file_records_under('games', 'Site', rs, b'dd'),
            None)


class Sqlite3bitapi_file_records_under(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('file_records_under')

    def tearDown(self):
        _delete_folder('file_records_under')

    def test_file_records_under_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'dbidentity'",
            self.api.file_records_under,
            *(None, None, None, None))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.file_records_under,
            *(None, None, rs, None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "file_records_under not implemented for Sqlite3bitPrimary",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.file_records_under,
            *('games', 'Site', rs, b'dd'))
        self.api.open_context()
        self.assertRaisesRegex(
            database.DatabaseError,
            "Record set was not created from this database instance",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            database.DatabaseError,
            "file_records_under not implemented for Sqlite3bitPrimary",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs._dbset = 'sssss'
        self.assertRaisesRegex(
            database.DatabaseError,
            "Record set was not created from dbset database",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertEqual(
            self.api.file_records_under('games', 'Site', rs, b'dd'),
            None)


class Sqlite3bitapi_method_classes(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_method_classes_01(self):
        bit = sqlite3api.Sqlite3bitapi
        api = sqlite3api._Sqlite3api
        base = database.Database
        enc = sqlite3api._DatabaseEncoders

        self.assertEqual(
            len([a for a in dir(bit)
                 if not a.startswith('__') and not a.endswith('__')]),
            50)
        self.assertEqual(
            len([a for a in dir(api)
                 if not a.startswith('__') and not a.endswith('__')]),
            50)
        self.assertEqual(
            len([a for a in dir(base)
                 if not a.startswith('__') and not a.endswith('__')]),
            25)
        self.assertEqual(
            len([a for a in dir(enc)
                 if not a.startswith('__') and not a.endswith('__')]),
            4)

        # Methods inherited from _Sqlite3api
        self.assertIs(api.backout, bit.backout)
        self.assertIs(api.close_contexts, bit.close_contexts)
        self.assertIs(api.close_database, bit.close_database)
        self.assertIs(api.commit, bit.commit)
        self.assertIs(api.db_compatibility_hack, bit.db_compatibility_hack)
        self.assertIs(api.exists, bit.exists)
        self.assertIs(api.files_exist, bit.files_exist)
        self.assertIs(api.database_cursor, bit.database_cursor)
        self.assertIs(api.repair_cursor, bit.repair_cursor)
        self.assertIs(api.get_database_folder, bit.get_database_folder)
        self.assertIs(api.get_database, bit.get_database)
        self.assertIs(api.get_database_instance, bit.get_database_instance)
        self.assertIs(
            api.get_first_primary_key_for_index_key,
            bit.get_first_primary_key_for_index_key)
        self.assertIs(api.get_primary_record, bit.get_primary_record)
        self.assertIs(api.is_primary, bit.is_primary)
        self.assertIs(api.is_primary_recno, bit.is_primary_recno)
        self.assertIs(api.is_recno, bit.is_recno)
        self.assertIs(api.open_contexts, bit.open_contexts)
        self.assertIs(api.allocate_and_open_contexts,
                      bit.allocate_and_open_contexts)
        self.assertIs(api.get_packed_key, bit.get_packed_key)
        self.assertIs(api.decode_as_primary_key, bit.decode_as_primary_key)
        self.assertIs(api.encode_primary_key, bit.encode_primary_key)
        self.assertIs(
            api.use_deferred_update_process,
            bit.use_deferred_update_process)
        self.assertIs(api.initial_database_size, bit.initial_database_size)
        self.assertIs(api.increase_database_size, bit.increase_database_size)
        self.assertIs(api.do_database_task, bit.do_database_task)
        self.assertIs(api.make_recordset_key, bit.make_recordset_key)
        self.assertIs(
            api.make_recordset_key_startswith,
            bit.make_recordset_key_startswith)
        self.assertIs(
            api.make_recordset_key_range,
            bit.make_recordset_key_range)
        self.assertIs(api.make_recordset_all, bit.make_recordset_all)
        self.assertIs(api.recordset_for_segment, bit.recordset_for_segment)
        self.assertIs(api.file_records_under, bit.file_records_under)
        self.assertIs(api.is_engine_uses_bytes, bit.is_engine_uses_bytes)
        self.assertIs(api.is_engine_uses_str, bit.is_engine_uses_str)
        self.assertIs(api.start_transaction, bit.start_transaction)
        self.assertIs(api.cede_contexts_to_process,
                      bit.cede_contexts_to_process)
        
        # Methods inherited from _DatabaseEncoders
        self.assertIs(enc.encode_record_number, bit.encode_record_number)
        self.assertIs(enc.decode_record_number, bit.decode_record_number)
        self.assertIs(enc.encode_record_selector, bit.encode_record_selector)
        self.assertIs(enc.encode_record_key, bit.encode_record_key)
        
        # Methods inherited from Database
        self.assertIs(base.deferred_update_housekeeping,
                      bit.deferred_update_housekeeping)

        # _Sqlite3api methods overridden
        self.assertIsNot(api.__init__, bit.__init__)
        self.assertIsNot(api.open_context, bit.open_context)
        self.assertIsNot(api.close_context, bit.close_context)
        self.assertIsNot(api.delete_instance, bit.delete_instance)
        self.assertIsNot(api.edit_instance, bit.edit_instance)
        self.assertIsNot(api.put_instance, bit.put_instance)
        self.assertIsNot(api.do_deferred_updates, bit.do_deferred_updates)
        self.assertIsNot(api.set_defer_update, bit.set_defer_update)
        self.assertIsNot(api.unset_defer_update, bit.unset_defer_update)


class Sqlite3bitapi_init(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 1 required positional argument: ",
                "'sqlite3tables'",
                )),
            sqlite3api.Sqlite3bitapi,
            )
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Database file name None is not valid",
            sqlite3api.Sqlite3bitapi,
            *(samplefilespec(),),
            **dict(sqlite3databasefolder=None)
            )
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Table definitions must be a dictionary",
            sqlite3api.Sqlite3bitapi,
            *(None,),
            **dict(sqlite3databasefolder='')
            )

    def test___init___02(self):
        fs = dict(
            t=dict(
                primary='key',
                fields=dict(lock={}, key={}),
                filedesc=dict(fileorg=None),
                secondary=dict(key='lock'),
                ))
        api = sqlite3api.Sqlite3bitapi(
            dict(
                t=dict(
                    primary='key',
                    fields=dict(lock={}, key={}),
                    filedesc=dict(fileorg=None),
                    secondary=dict(key='lock'),
                    )),
            sqlite3databasefolder='')
        p = os.path.abspath('')
        self.assertEqual(api._sqfile, os.path.join(p, os.path.basename(p)))
        self.assertEqual(api._sqconn, None)
        self.assertEqual(api._associate, dict(t=dict(key='lock', t='key')))
        self.assertEqual(len(api._sqtables), 2)
        self.assertIsInstance(
            api._sqtables['key'], sqlite3api.Sqlite3bitPrimary)
        self.assertIsInstance(
            api._sqtables['lock'], sqlite3api.Sqlite3bitSecondary)
        self.assertIsInstance(api._control, sqlite3api.Sqlite3bitControlFile)
        self.assertEqual(len(api.__dict__), 5)
        self.assertIsInstance(
            sqlite3api.Sqlite3bitapi(samplefilespec(),
                                     sqlite3databasefolder=''),
            sqlite3api.Sqlite3bitapi)
        self.assertIsInstance(
            sqlite3api.Sqlite3bitapi(samplefilespec(), ''),
            sqlite3api.Sqlite3bitapi)


class Sqlite3bitapi_close_context(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('close_context')

    def tearDown(self):
        _delete_folder('close_context')

    def test_close_context_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.close_context(), None)
        self.assertEqual(self.api._sqconn, None)

    def test_close_context_02(self):
        self.api.open_context()
        conn = self.api._sqconn
        self.assertIsInstance(conn, sqlite3.Connection)
        self.assertEqual(self.api.close_context(), None)
        self.assertIs(self.api._sqconn, conn)
        self.assertEqual(self.api.close_context(), None)
        self.assertIs(self.api._sqconn, conn)


class Sqlite3bitapi_open_context(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('open_context')

    def tearDown(self):
        _delete_folder('open_context')

    def test_open_context_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(os.path.exists(self.api._sqfile), False)
        self.assertEqual(self.api.open_context(), True)
        conn = self.api._sqconn
        self.assertIsInstance(conn, sqlite3.Connection)
        self.assertEqual(os.path.isfile(self.api._sqfile), True)
        self.assertEqual(self.api.open_context(), True)
        self.assertIs(self.api._sqconn, conn)
        self.assertEqual(os.path.isfile(self.api._sqfile), True)


class Sqlite3bitapi_set_defer_update(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('set_defer_update')

    def tearDown(self):
        _delete_folder('set_defer_update')

    def test_set_defer_update_01(self):
        self.assertEqual(self.api.set_defer_update(), False)
        self.assertEqual(self.api.set_defer_update(duallowed=True), True)
        self.assertEqual(self.api.set_defer_update(duallowed='xx'), 'xx')
        self.assertEqual(
            self.api.set_defer_update(db='Berkeley DB compatibility'), False)


class Sqlite3bitapi_unset_defer_update(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('unset_defer_update')

    def tearDown(self):
        _delete_folder('unset_defer_update')

    def test_unset_defer_update_01(self):
        self.assertEqual(
            self.api.unset_defer_update(db='Berkeley DB compatibility'), True)


class Sqlite3bitapi_do_deferred_updates(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('do_deferred_updates')

    def tearDown(self):
        _delete_folder('do_deferred_updates')

    def test_do_deferred_updates_01(self):
        self.api.open_context()
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "'pyscript' is not an existing file",
            self.api.do_deferred_updates,
            *('pyscript', 'filepath'))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "'filepath' is not an existing file",
            self.api.do_deferred_updates,
            *(self.api._sqfile, 'filepath'))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "'filepath' is not an existing file",
            self.api.do_deferred_updates,
            *(self.api._sqfile, (self.api._sqfile, 'filepath')))
        script = os.path.join(os.path.dirname(self.api._sqfile), 'script')
        open(script, 'w').close()
        sp = self.api.do_deferred_updates(script, self.api._sqfile)
        self.assertIsInstance(sp, subprocess.Popen)
        sp.wait()


class Sqlite3bitapi_put_instance(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('put_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('put_instance')

    def test_put_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.put_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 0
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.put_instance,
            *(None, instance))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 0
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.put_instance,
            *('games', instance))

    def test_put_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 0
        self.api.open_context()
        self.assertEqual(self.api.put_instance('games', instance), None)

    def test_put_instance_03(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 4
        self.api.open_context()
        self.assertEqual(self.api.put_instance('games', instance), None)
        self.assertEqual(self.api.put_instance('games', instance), None)

    def test_put_instance_04(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 0
        self.api.open_context()
        self.assertEqual(self.api.put_instance('games', instance), None)

    def test_put_instance_05(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=1,
                                    sgGame=0,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1,) is the record number.
                 records=dict(White=(('A Jones', 0, 1, (1,)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__exist=(b''.join((b'\x40', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 1, (1,)),),
                              ),
                 ))

    def test_put_instance_06(self):
        # commit inside record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(2):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
            self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=2,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used.
                 # White will get one and Black the other, at random.
                 records=dict(White=(('A Jones', 0, 2, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                 for i in range(1, 3)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 3)])),
                              Game__exist=(b''.join((b'\x60', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 2, (1, 2)),),
                              ),
                 ))

    def test_put_instance_07(self):
        # commit inside record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
            self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=3,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 3, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 4)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 4)])),
                              Game__exist=(b''.join((b'\x70', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 3, (1, 2)),),
                              ),
                 ))

    def test_put_instance_08(self):
        # commit inside record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(200):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
            self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=200,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 200, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 201)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 201)])),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*24,
                                                     b'\x80',
                                                     b'\x00'*8166)),),
                              Black=(('B Smith', 0, 200, (1, 2)),),
                              ),
                 ))

    def test_put_instance_09(self):
        # commit after record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(2000):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=2000,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 2000, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*249,
                                                b'\x80',
                                                b'\x00'*7941)),
                                      b''.join((b'\x7f',
                                                b'\xff'*249,
                                                b'\x80',
                                                b'\x00'*7941)),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*249,
                                                     b'\x80',
                                                     b'\x00'*7941)),
                                           ),
                              Black=(('B Smith', 0, 2000, (1, 2)),),
                              ),
                 ))

    def test_put_instance_10(self):
        # commit after record add loop
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.api.open_context()
        for i in range(66000):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66000,
                                    sgGame=4,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 465, (3, 4)),
                                     ),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                     b'\xff'*8191)),
                                      b''.join((b'\x7f',
                                                b'\xff'*8191)),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(465)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(465)]),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*8191)),
                                           b''.join((b'\xff'*58,
                                                     b'\x80',
                                                     b'\x00'*8133)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 465, (3, 4)),
                                     ),
                              ),
                 ))


class Sqlite3bitapi_delete_instance(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('delete_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('delete_instance')

    def test_delete_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.delete_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.delete_instance,
            *(None, instance))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.delete_instance,
            *('games', instance))

    def test_delete_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.api.open_context()
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Existence bit map for segment does not exist",
            self.api.delete_instance,
            *('games', instance))

    def test_delete_instance_03(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 0
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 100000
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Existence bit map for segment does not exist",
            self.api.delete_instance,
            *('games', instance))
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 10000
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 1
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 1
        self.assertEqual(self.api.delete_instance('games', instance), None)

    def test_delete_instance_04(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # Do a put_instance test to establish starting point.
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=3,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 3, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 4)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 4)])),
                              Game__exist=(b''.join((b'\x70', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 3, (1, 2)),),
                              ),
                 ))
        # Do the delete_instance test.
        # Delete record number 2
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 2))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=3,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 2, (1, 2)),),
                              Game=dict(null={2},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                for i in (1, 3)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in (1, 3)])),
                              Game__exist=(b''.join((b'\x50', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 2, (1, 2)),),
                              ),
                 ))
        # Delete record number 3
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 3))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=3,
                                    sgGame=0,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 1, (1, 2)),),
                              Game=dict(null={2, 3},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              Game__exist=(b''.join((b'\x40', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 1, (1, 2)),),
                              ),
                 ))

    def test_delete_instance_05(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # Do a put_instance test to establish starting point.
        # Exactly the number of records so second delete from segment number 1
        # should convert segment bitmap to list of record numbers:
        # range(65536 + DB_CONVERSION_LIMIT + 1).
        self.api.open_context()
        for i in range(66048):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66048,
                                    sgGame=4,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 513, (3, 4)),
                                     ),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                     b'\xff'*8191)),
                                      b''.join((b'\x7f',
                                                b'\xff'*8191)),
                                      b''.join((b'\xff'*64,
                                                b'\x80',
                                                b'\x00'*8127)),
                                      b''.join((b'\xff'*64,
                                                b'\x80',
                                                b'\x00'*8127)),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*8191)),
                                           b''.join((b'\xff'*64,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 513, (3, 4)),
                                     ),
                              ),
                 ))
        # Do the delete_instance test.
        # Delete record number 58
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 58))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66048,
                                    sgGame=4,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=1,
                                    ),
                 records=dict(White=(('A Jones', 0, 65534, (1, 2)),
                                     ('A Jones', 1, 513, (3, 4)),
                                     ),
                              Game=dict(null={58},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xdf',
                                                b'\xff'*8184)),
                                      b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xdf',
                                                b'\xff'*8184)),
                                      b''.join((b'\xff'*31,
                                                b'\xff',
                                                b'\xff'*32,
                                                b'\x80',
                                                b'\x00'*8127)),
                                      b''.join((b'\xff'*31,
                                                b'\xff',
                                                b'\xff'*32,
                                                b'\x80',
                                                b'\x00'*8127)),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xdf',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xff',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65534, (1, 2)),
                                     ('B Smith', 1, 513, (3, 4)),
                                     ),
                              ),
                 ))
        # Delete record number 65790
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65790))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66048,
                                    sgGame=4,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=1,
                                    ),
                 records=dict(White=(('A Jones', 0, 65534, (1, 2)),
                                     ('A Jones', 1, 512, (3, 4)),
                                     ),
                              Game=dict(null={58, 65790},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xdf',
                                                b'\xff'*8184)),
                                      b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xdf',
                                                b'\xff'*8184)),
                                      b''.join((b'\xff'*31,
                                                b'\xfd',
                                                b'\xff'*32,
                                                b'\x80',
                                                b'\x00'*8127)),
                                      b''.join((b'\xff'*31,
                                                b'\xfd',
                                                b'\xff'*32,
                                                b'\x80',
                                                b'\x00'*8127)),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xdf',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xfd',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65534, (1, 2)),
                                     ('B Smith', 1, 512, (3, 4)),
                                     ),
                              ___control=(('E', 0),
                                          ),
                              ),
                 ))
        # Delete record number 65787
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65787))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66048,
                                    sgGame=4,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=1,
                                    ),
                 records=dict(White=(('A Jones', 0, 65534, (1, 2)),
                                     ('A Jones', 1, 511, (3, 4)),
                                     ),
                              Game=dict(null={58, 65790, 65787},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xdf',
                                                b'\xff'*8184)),
                                      b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xdf',
                                                b'\xff'*8184)),
                                      b''.join((
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(31*8)]),
                                          b''.join([(i+31*8).to_bytes(2, 'big')
                                                    for i in
                                                    (0, 1, 2, 4, 5, 7)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(
                                                        32*8, 1+64*8)]),
                                          )),
                                      b''.join((
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(31*8)]),
                                          b''.join([(i+31*8).to_bytes(2, 'big')
                                                    for i in
                                                    (0, 1, 2, 4, 5, 7)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(
                                                        32*8, 1+64*8)]),
                                          )),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xdf',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65534, (1, 2)),
                                     ('B Smith', 1, 511, (3, 4)),
                                     ),
                              ___control=(('E', 0),
                                          ),
                              ),
                 ))


class Sqlite3bitapi_edit_instance(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('edit_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('edit_instance')

    def test_edit_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.edit_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, record.Value)
        edited.key.recno = 2
        instance.newrecord = edited
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.edit_instance,
            *(None, instance))

    def test_edit_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, record.Value)
        edited.key.recno = 2
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_03(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 2
        instance.newrecord = edited
        self.api.open_context()
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_04(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 2
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_05(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 3
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_06(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        editvalue = {WHITE_FIELD_DEF:'C Brown',
                     BLACK_FIELD_DEF:'D Grey',
                     }
        # set starting point with 3 records.
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=3,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used.
                 # White will get one and Black the other, at random.
                 records=dict(White=(('A Jones', 0, 3, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                 for i in range(1, 4)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 4)])),
                              Game__exist=(b''.join((b'\x70', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 3, (1, 2)),),
                              ),
                 ))
        self.api.open_context()
        # Edit record number 2.
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 2))
        er = self.sr.clone()
        er.value.White = editvalue[WHITE_FIELD_DEF]
        er.value.Black = editvalue[BLACK_FIELD_DEF]
        self.sr.newrecord = er
        self.api.edit_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=3,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used.
                 # White will get one and Black the other, at random.
                 # Brown and Grey are direct record numbers (one record only).
                 records=dict(White=(('A Jones', 0, 2, (1, 2)),
                                     ('C Brown', 0, 1, (2,)),
                                     ),
                              Game=dict(null={},
                                        rows={1:dict(
                                            White='C Brown', Black='D Grey') },
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                 for i in (1, 3)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in (1, 3)])),
                              Game__exist=(b''.join((b'\x70', b'\x00'*8191)),),
                              Black=(('B Smith', 0, 2, (1, 2)),
                                     ('D Grey', 0, 1, (2,)),
                                     ),
                              ),
                 ))

    def test_edit_instance_07(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        editvalue = {WHITE_FIELD_DEF:'C Brown',
                     BLACK_FIELD_DEF:'D Grey',
                     }
        # set starting point with 9 records.
        # To get lists of record numbers changed by third edit.
        self.api.open_context()
        for i in range(9):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=1,
                                    Game=9,
                                    sgGame=2,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=1,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used.
                 # White will get one and Black the other, at random.
                 records=dict(White=(('A Jones', 0, 9, (1, 2)),),
                              Game=dict(null={},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                 for i in range(1, 10)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in range(1, 10)])),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xc0',
                                                     b'\x00'*8190)),),
                              Black=(('B Smith', 0, 9, (1, 2)),),
                              ),
                 ))
        self.api.open_context()
        # Edit record numbers 2, 5, 6.
        for i in (2, 5, 6):
            self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, i))
            er = self.sr.clone()
            er.value.White = editvalue[WHITE_FIELD_DEF]
            er.value.Black = editvalue[BLACK_FIELD_DEF]
            self.sr.newrecord = er
            self.api.edit_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=9,
                                    sgGame=4,
                                    Game__exist=1,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 # (1, 2) is the segment record numbers which could be used,
                 # White will get one and Black the other, at random.
                 # Brown and Grey get segment numbers (3, 4).
                 records=dict(White=(('A Jones', 0, 6, (1, 2)),
                                     ('C Brown', 0, 3, (3, 4)),
                                     ),
                              Game=dict(null={},
                                        rows={k:dict(White='C Brown',
                                                     Black='D Grey')
                                              for k in (1, 4, 5)},
                                        default=dict(White='A Jones',
                                                     Black='B Smith'),
                                        ),
                              sgGame=(b''.join([i.to_bytes(2, 'big')
                                                 for i in (1, 3, 4, 7, 8, 9)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                 for i in (1, 3, 4, 7, 8, 9)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                 for i in (2, 5, 6)]),
                                      b''.join([i.to_bytes(2, 'big')
                                                for i in (2, 5, 6)]),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xc0',
                                                     b'\x00'*8190)),),
                              Black=(('B Smith', 0, 6, (1, 2)),
                                     ('D Grey', 0, 3, (3, 4)),
                                     ),
                              ),
                 ))


class Sqlite3bitapi_put_instance_with_holes(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('put_instance_holes')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('put_instance_holes')

    def test_holes_01(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # set starting point like end of test_delete_instance_05
        self.api.open_context()
        for i in range(66048):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 58))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65790))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65787))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66048,
                                    sgGame=4,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=1,
                                    ),
                 records=dict(White=(('A Jones', 0, 65534, (1, 2)),
                                     ('A Jones', 1, 511, (3, 4)),
                                     ),
                              Game=dict(null={58, 65790, 65787},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xdf',
                                                b'\xff'*8184)),
                                      b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xdf',
                                                b'\xff'*8184)),
                                      b''.join((
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(31*8)]),
                                          b''.join([(i+31*8).to_bytes(2, 'big')
                                                    for i in
                                                    (0, 1, 2, 4, 5, 7)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(
                                                        32*8, 1+64*8)]),
                                          )),
                                      b''.join((
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(31*8)]),
                                          b''.join([(i+31*8).to_bytes(2, 'big')
                                                    for i in
                                                    (0, 1, 2, 4, 5, 7)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(
                                                        32*8, 1+64*8)]),
                                          )),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xdf',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65534, (1, 2)),
                                     ('B Smith', 1, 511, (3, 4)),
                                     ),
                              ___control=(('E', 0),
                                          ),
                              ),
                 ))
        # Insert a record
        self.api.open_context()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=66048,
                                    sgGame=4,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=1,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 511, (3, 4)),
                                     ),
                              Game=dict(null={65790, 65787},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xff',#df',
                                                b'\xff'*8184)),
                                      b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xff',#df',
                                                b'\xff'*8184)),
                                      b''.join((
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(31*8)]),
                                          b''.join([(i+31*8).to_bytes(2, 'big')
                                                    for i in
                                                    (0, 1, 2, 4, 5, 7)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(
                                                        32*8, 1+64*8)]),
                                          )),
                                      b''.join((
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(31*8)]),
                                          b''.join([(i+31*8).to_bytes(2, 'big')
                                                    for i in
                                                    (0, 1, 2, 4, 5, 7)]),
                                          b''.join([i.to_bytes(2, 'big')
                                                    for i in range(
                                                        32*8, 1+64*8)]),
                                          )),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xff',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\x80',
                                                     b'\x00'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 511, (3, 4)),
                                     ),
                              ___control=(('E', 0),
                                          ),
                              ),
                 ))
        # Fill the high segment
        self.api.open_context()
        for i in range(1+64*8, 65536):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=131071,
                                    sgGame=4,
                                    Game__exist=2,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=0,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 65534, (3, 4)),
                                     ),
                              Game=dict(null={65790, 65787},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xff',
                                                b'\xff'*8184)),
                                      b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xff',
                                                b'\xff'*8184)),
                                      # Segment 1 back to bitmap form.
                                      b''.join((b'\xff'*31,
                                                b'\xed',
                                                b'\xff'*32,
                                                b'\xff',
                                                b'\xff'*8127)),
                                      b''.join((b'\xff'*31,
                                                b'\xed',
                                                b'\xff'*32,
                                                b'\xff',
                                                b'\xff'*8127)),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xff',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\xff',
                                                     b'\xff'*8127)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 65534, (3, 4)),
                                     ),
                              #___control=(('E', 0),
                              #            ),
                              ),
                 ))
        # Delete a record to set reuse for segment 1.
        self.api.open_context()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 66053))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 131072))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=131072,
                                    sgGame=4,
                                    Game__exist=3,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=1,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 65533, (3, 4)),
                                     ),
                              Game=dict(null={65790, 65787, 66053, 131072},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xff',
                                                b'\xff'*8184)),
                                      b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xff',
                                                b'\xff'*8184)),
                                      b''.join((b'\xff'*31,
                                                b'\xed',
                                                b'\xff'*32,
                                                b'\xfb',
                                                b'\xff'*8127)),
                                      b''.join((b'\xff'*31,
                                                b'\xed',
                                                b'\xff'*32,
                                                b'\xfb',
                                                b'\xff'*8127)),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xff',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xed',
                                                     b'\xff'*32,
                                                     b'\xfb',
                                                     b'\xff'*8127)),
                                           b''.join((b'\x00',
                                                     b'\x00'*8191)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 65533, (3, 4)),
                                     ),
                              ___control=(('E', 1),
                                          ),
                              ),
                 ))
        # Insert three records.
        self.api.open_context()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.sr.load_record((0, repr(value)))
        self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        record_counts(
            self,
            self.api.get_database_folder(),
            dict(record_counts=dict(Round=0,
                                    White=2,
                                    Game=131072,
                                    sgGame=4,
                                    Game__exist=3,
                                    Event=0,
                                    Site=0,
                                    Black=2,
                                    Date=0,
                                    Result=0,
                                    ___control=1,
                                    ),
                 records=dict(White=(('A Jones', 0, 65535, (1, 2)),
                                     ('A Jones', 1, 65536, (3, 4)),
                                     ),
                              Game=dict(null={131072},
                                        rows=dict(),
                                        default=dict(
                                            White='A Jones', Black='B Smith'),
                                        ),
                              sgGame=(b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xff',
                                                b'\xff'*8184)),
                                      b''.join((b'\x7f',
                                                b'\xff'*6,
                                                b'\xff',
                                                b'\xff'*8184)),
                                      b''.join((b'\xff'*31,
                                                b'\xff',
                                                b'\xff'*32,
                                                b'\xff',
                                                b'\xff'*8127)),
                                      b''.join((b'\xff'*31,
                                                b'\xff',
                                                b'\xff'*32,
                                                b'\xff',
                                                b'\xff'*8127)),
                                      ),
                              Game__exist=(b''.join((b'\x7f',
                                                     b'\xff'*6,
                                                     b'\xff',
                                                     b'\xff'*8184)),
                                           b''.join((b'\xff'*31,
                                                     b'\xff',
                                                     b'\xff'*32,
                                                     b'\xff',
                                                     b'\xff'*8127)),
                                           b''.join((b'\x00',
                                                     b'\x00'*8191)),
                                           ),
                              Black=(('B Smith', 0, 65535, (1, 2)),
                                     ('B Smith', 1, 65536, (3, 4)),
                                     ),
                              ___control=(('E', 1),
                                          ),
                              ),
                 ))


class Sqlite3api_delete_instance_no_counting(unittest.TestCase):

    def setUp(self):
        self.api = api('delete_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('delete_instance')

    def test_delete_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.delete_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.delete_instance,
            *(None, instance))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.assertEqual(self.api.delete_instance('games', instance), None)

    def test_delete_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.api.open_context()
        self.assertEqual(self.api.delete_instance('games', instance), None)

    def test_delete_instance_03(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 0
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 100000
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 10000
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 1
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 1
        self.assertEqual(self.api.delete_instance('games', instance), None)

    def test_delete_instance_04(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # Do a put_instance test to establish starting point.
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Do the delete_instance test.
        # Delete record number 2
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 2))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Delete record number 3
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 3))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()

    def test_delete_instance_05(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # Do a put_instance test to establish starting point.
        # Exactly the number of records so second delete from segment number 1
        # should convert segment bitmap to list of record numbers:
        # range(65536 + DB_CONVERSION_LIMIT + 1).
        self.api.open_context()
        for i in range(66048):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Do the delete_instance test.
        # Delete record number 58
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 58))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Delete record number 65790
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65790))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Delete record number 65787
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65787))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()


class Sqlite3api_edit_instance_no_counting(unittest.TestCase):

    def setUp(self):
        self.api = api('edit_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('edit_instance')

    def test_edit_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.edit_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, record.Value)
        edited.key.recno = 2
        instance.newrecord = edited
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.edit_instance,
            *(None, instance))

    def test_edit_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, record.Value)
        edited.key.recno = 2
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_03(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 2
        instance.newrecord = edited
        self.api.open_context()
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_04(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 2
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_05(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 3
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_06(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        editvalue = {WHITE_FIELD_DEF:'C Brown',
                     BLACK_FIELD_DEF:'D Grey',
                     }
        # set starting point with 3 records.
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        self.api.open_context()
        # Edit record number 2.
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 2))
        er = self.sr.clone()
        er.value.White = editvalue[WHITE_FIELD_DEF]
        er.value.Black = editvalue[BLACK_FIELD_DEF]
        self.sr.newrecord = er
        self.api.edit_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()

    def test_edit_instance_07(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        editvalue = {WHITE_FIELD_DEF:'C Brown',
                     BLACK_FIELD_DEF:'D Grey',
                     }
        # set starting point with 9 records.
        # To get lists of record numbers changed by third edit.
        self.api.open_context()
        for i in range(9):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        self.api.open_context()
        # Edit record numbers 2, 5, 6.
        for i in (2, 5, 6):
            self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, i))
            er = self.sr.clone()
            er.value.White = editvalue[WHITE_FIELD_DEF]
            er.value.Black = editvalue[BLACK_FIELD_DEF]
            self.sr.newrecord = er
            self.api.edit_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()


class Sqlite3bitapi_delete_instance_no_counting(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('delete_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('delete_instance')

    def test_delete_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.delete_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.delete_instance,
            *(None, instance))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'execute'",
            self.api.delete_instance,
            *('games', instance))

    def test_delete_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.api.open_context()
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Existence bit map for segment does not exist",
            self.api.delete_instance,
            *('games', instance))

    def test_delete_instance_03(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 0
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 100000
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Existence bit map for segment does not exist",
            self.api.delete_instance,
            *('games', instance))
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 10000
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 1
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 1
        self.assertEqual(self.api.delete_instance('games', instance), None)

    def test_delete_instance_04(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # Do a put_instance test to establish starting point.
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Do the delete_instance test.
        # Delete record number 2
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 2))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Delete record number 3
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 3))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()

    def test_delete_instance_05(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        # Do a put_instance test to establish starting point.
        # Exactly the number of records so second delete from segment number 1
        # should convert segment bitmap to list of record numbers:
        # range(65536 + DB_CONVERSION_LIMIT + 1).
        self.api.open_context()
        for i in range(66048):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Do the delete_instance test.
        # Delete record number 58
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 58))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Delete record number 65790
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65790))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        # Delete record number 65787
        self.api.open_context()
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 65787))
        self.api.delete_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()


class Sqlite3bitapi_edit_instance_no_counting(unittest.TestCase):

    def setUp(self):
        self.api = bitapi('edit_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        _delete_folder('edit_instance')

    def test_edit_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.edit_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, record.Value)
        edited.key.recno = 2
        instance.newrecord = edited
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.edit_instance,
            *(None, instance))

    def test_edit_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, record.Value)
        edited.key.recno = 2
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_03(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 2
        instance.newrecord = edited
        self.api.open_context()
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_04(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 2
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_05(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 3
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_06(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        editvalue = {WHITE_FIELD_DEF:'C Brown',
                     BLACK_FIELD_DEF:'D Grey',
                     }
        # set starting point with 3 records.
        self.api.open_context()
        for i in range(3):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        self.api.open_context()
        # Edit record number 2.
        self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, 2))
        er = self.sr.clone()
        er.value.White = editvalue[WHITE_FIELD_DEF]
        er.value.Black = editvalue[BLACK_FIELD_DEF]
        self.sr.newrecord = er
        self.api.edit_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()

    def test_edit_instance_07(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        editvalue = {WHITE_FIELD_DEF:'C Brown',
                     BLACK_FIELD_DEF:'D Grey',
                     }
        # set starting point with 9 records.
        # To get lists of record numbers changed by third edit.
        self.api.open_context()
        for i in range(9):
            self.sr.load_record((0, repr(value)))
            self.api.put_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()
        self.api.open_context()
        # Edit record numbers 2, 5, 6.
        for i in (2, 5, 6):
            self.sr.load_record(self.api.get_primary_record(GAMES_FILE_DEF, i))
            er = self.sr.clone()
            er.value.White = editvalue[WHITE_FIELD_DEF]
            er.value.Black = editvalue[BLACK_FIELD_DEF]
            self.sr.newrecord = er
            self.api.edit_instance(GAMES_FILE_DEF, self.sr)
        self.api.commit()
        self.api.close_context()
        no_record_counts()


def no_record_counts(*a):
    pass


def record_counts(testcase, db_directory, answers):
    literal_eval = ast.literal_eval
    testcase.assertEqual(
        bool(set(answers['records']) - set(answers['record_counts'])),
        False)
    dbfile = os.path.join(db_directory, os.path.split(db_directory)[-1])
    conn = sqlite3.connect(dbfile)
    cursor = conn.cursor()
    for t in ('Round',
              'White',
              'Game',
              'sgGame',
              'Game__exist',
              'Event',
              'Site',
              'Black',
              'Date',
              'Result',
              '___control',
              ):
        cursor.execute(t.join(('select count(*) from ', '')))
        testcase.assertEqual(cursor.fetchone()[0], answers['record_counts'][t])
    segments = set()
    for k, v in answers['records'].items():
        cursor.execute(k.join(('select * from ', '')))
        # Sqlite3 database rows are 1-base but tuple indexes are 0-base.
        row = 0
        while True:
            r = cursor.fetchone()
            if not r:
                break
            row += 1
            if k == 'Game':
                # An empty row exists for deleted records in sqlite3 but not
                # in bsddb3 databases. This allows the segment slot for a
                # deleted record to be reused.
                if r[0] in v['null']:
                    testcase.assertEqual(r[1], None)
                else:
                    testcase.assertEqual(r[0], row)
                    testcase.assertEqual(
                        literal_eval(r[1]),
                        v['rows'].get(row-1, v['default']))
            elif k in ('White', 'Black'):
                # r[-1] is direct record number if r[2], count, is 1.
                if r[2] > 1:
                    segments.add(r[-1])
                testcase.assertEqual(r[:-1], v[row-1][:-1])
                # Cannot be sure which value got used, but 'White' and
                # 'Black' must use a different one of those given.
                testcase.assertEqual(r[-1] in v[row-1][-1], True)
            elif k == 'Game__exist':
                testcase.assertEqual(r[0], row)
                testcase.assertEqual(r[-1], v[row-1])
            elif k == 'sgGame':
                testcase.assertEqual(r[-1], v[row-1])
            elif k == '___control':
                # Queue of segments, other than the high segment, with unused
                # record numbers after record deletion.
                testcase.assertEqual(r[0].decode(), v[row-1][0])
                testcase.assertEqual(r[1], v[row-1][-1])
    # An odd number of segments means 'White' and 'Black' shared a segment.
    testcase.assertEqual(len(segments) % 2, 0)
    del cursor
    conn.close()


if __name__ == '__main__':
    runner = unittest.TextTestRunner
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    
    runner().run(loader(_DatabaseEncoders))

    # The _Sqlite3api class tests.
    
    runner().run(loader(_Sqlite3api_init))

    # These three not in definition order.
    runner().run(loader(_Sqlite3api_open_context))
    runner().run(loader(_Sqlite3api_close_context))
    runner().run(loader(_Sqlite3api_close_database))

    # The rest in definition order.
    runner().run(loader(_Sqlite3api_backout))
    runner().run(loader(_Sqlite3api_close_contexts))
    runner().run(loader(_Sqlite3api_commit))
    runner().run(loader(_Sqlite3api_db_compatibility_hack))
    runner().run(loader(_Sqlite3api_exists))
    runner().run(loader(_Sqlite3api_files_exist))
    runner().run(loader(_Sqlite3api_make_cursor))
    runner().run(loader(_Sqlite3api_get_database_folder))
    runner().run(loader(_Sqlite3api_get_database))
    runner().run(loader(_Sqlite3api_get_database_instance))
    runner().run(loader(_Sqlite3api_get_first_primary_key_for_index_key))
    runner().run(loader(_Sqlite3api_get_primary_record))
    runner().run(loader(_Sqlite3api_is_primary))
    runner().run(loader(_Sqlite3api_is_primary_recno))
    runner().run(loader(_Sqlite3api_is_recno))
    runner().run(loader(_Sqlite3api_open_contexts))
    runner().run(loader(_Sqlite3api_get_packed_key))
    runner().run(loader(_Sqlite3api_decode_as_primary_key))
    runner().run(loader(_Sqlite3api_encode_primary_key))
    runner().run(loader(_Sqlite3api_use_deferred_update_process))
    runner().run(loader(_Sqlite3api_initial_database_size))
    runner().run(loader(_Sqlite3api_increase_database_size))
    runner().run(loader(_Sqlite3api_do_database_task))
    runner().run(loader(_Sqlite3api_make_recordset_key))
    runner().run(loader(_Sqlite3api_make_recordset_key_startswith))
    runner().run(loader(_Sqlite3api_make_recordset_key_range))
    runner().run(loader(_Sqlite3api_make_recordset_all))
    runner().run(loader(_Sqlite3api_recordset_for_segment))
    runner().run(loader(_Sqlite3api_file_records_under))

    # The Sqlite3bitapi class tests.
    
    runner().run(loader(Sqlite3bitapi_method_classes))
    runner().run(loader(Sqlite3bitapi_init))
    runner().run(loader(Sqlite3bitapi_open_context))
    runner().run(loader(Sqlite3bitapi_close_context))
    runner().run(loader(Sqlite3bitapi_do_database_task))
    runner().run(loader(Sqlite3bitapi_make_recordset_key_startswith))
    runner().run(loader(Sqlite3bitapi_recordset_for_segment))
    runner().run(loader(Sqlite3bitapi_file_records_under))
    runner().run(loader(Sqlite3bitapi_set_defer_update))
    runner().run(loader(Sqlite3bitapi_unset_defer_update))
    runner().run(loader(Sqlite3bitapi_do_deferred_updates))
    runner().run(loader(Sqlite3bitapi_put_instance))
    runner().run(loader(Sqlite3bitapi_delete_instance))
    runner().run(loader(Sqlite3bitapi_edit_instance))

    # Some important sequences
    
    runner().run(loader(Sqlite3bitapi_put_instance_with_holes))

    # Critical Sqlite3api class tests compared with Sqlite3bitapi versions.

    # At 17 March 2016 on FreeBSD 10.1 using portsnap downloaded on 7 February
    # 2016 the following 'Sqlite3api_*' tests are OK for Python 3.3, 3.4, and
    # 3.5; while the 'Sqlite3bitapi_*' tests are OK for Python 3.3 but give
    # 'sqlite3.OperationalError: disk I/O error' for Python 3.4 and 3.5 in
    # most runs.

    runner().run(loader(Sqlite3api_delete_instance_no_counting))
    runner().run(loader(Sqlite3bitapi_delete_instance_no_counting))
    runner().run(loader(Sqlite3api_edit_instance_no_counting))
    runner().run(loader(Sqlite3bitapi_edit_instance_no_counting))
